from enums.EnumAsClass import EnumAsClass


class TimerKeys(EnumAsClass):
    TOTAL_TIME = "total_time"
    EXTRACT_TIME = "extract_time"
    TRANSFORM_TIME = "transform_time"
    LOAD_TIME = "load_time"
    STATISTICS_TIME = "statistics_time"
    REPORT_TIME = "report_time"
    INSERT_DATASETS = "insert_datasets"
    PROFILE_COMPUTATION = "profile_computation"
    READ_TABULAR = "read_tabular"
    PREPROCESS_DATA = "preprocess_data"
    NORMALIZATION_DATA = "normalization_data"
    NORMALIZATION_METADATA = "normalization_metadata"
    FILTER_DATA = "filter_data"
    FILTER_METADATA = "filter_metadata"
    COMPUTE_MAPPINGS_OR = "compute_mappings_or"
    COMPUTE_MAPPINGS_UNIT = "compute_mappings_domain"
    COMPUTE_MAPPINGS_DOMAIN = "compute_mappings_unit"
    OR_CREATION_TIME = "ontology_resource_creation_time"
    API_CALLS_TIME = "api_calls_time"
    API_CALL_USER_AGENT = "API_CALL_USER_AGENT"
    API_CALL_AUTHENTICATION = "API_CALL_AUTHENTICATION"
    API_CALL_KEY_HEADER = "API_CALL_KEY_HEADER"
    API_CALL_KEY_BEARER = "API_CALL_KEY_BEARER"
    API_CALL_KEY_URL = "API_CALL_KEY_URL"
    API_CALL_SECOND_TRY = "api_call_second_try"
    API_PROCESS_RESPONSE = "api_process_response"
    DB_WRITE_TIME = "db_write_time"
    LOAD_JSON_IN_MONGO = "load_json_in_mongo"
    UPSERT_TUPLES = "upsert_tuples"

    # # Transform
    CREATE_PATIENTS = "create_patients"
    WRITE_PATIENTS_IN_DB = "write_patients_in_db"
    CREATE_FEATURES = "create_features"
    CREATE_FEATURE_INSTANCES = "create_feature_instances"
    PROCESS_FEATURE_BATCH = "process_feature_batch"
    WRITE_FEATURES_IN_DB = "write_features_in_db"
    CREATE_RECORDS = "create_records"
    CREATE_RECORD_INSTANCES = "create_record_instances"
    GET_IDS = "GET_IDS"
    IF_ENTITY_TYPE = "IF_ENTITY_TYPE"
    PROCESS_RECORD_BATCH = "process_record_batch"
    WRITE_RECORDS_IN_DB = "write_records_in_db"

    # # Fairification
    VALUE_FAIRIFICATION = "value_fairification"

    # Writing record resources
    GET_RESOURCE_JSON_FILE = "get_resource_json_file"
    JSONIFY_RECORDS = "jsonify_records"
    UJSON_DUMP = "ujson_dump"
    CREATE_UPSERT_OPERATIONS = "CREATE_UPSERT_OPERATIONS"
    BULK_WRITE = "BULK_WRITE"

    # increase file size to reach the 16Mb limit
    WRITE_IN_FILE_FEATURES = "write_in_file_features"
    WRITE_IN_FILE_RECORDS = "write_in_file_records"
    WRITE_IN_FILE_PATIENTS = "write_in_file_patients"
    CHECK_BATCH_SIZE_FEATURES = "check_batch_size_features"
    CHECK_BATCH_SIZE_RECORDS = "check_batch_size_records"
    CHECK_BATCH_SIZE_PATIENTS = "check_batch_size_patients"



class EnumAsClass:
    @classmethod
    def values(cls):
        xs = []
        for name, value in vars(cls).items():
            if not (name.startswith('__') or isinstance(value, classmethod)):
                xs.append(value)
        return xs



from __future__ import annotations

import os

from constants.structure import DOCKER_FOLDER_TEST, DOCKER_FOLDER_METADATA, \
    DOCKER_FOLDER_DATA
from enums.EnumAsClass import EnumAsClass
from enums.ParameterKeys import ParameterKeys


class Profile(EnumAsClass):
    PHENOTYPIC = "phenotypic"
    CLINICAL = "clinical"
    DIAGNOSIS = "diagnosis"
    MEDICINE = "medicine"
    GENOMIC = "genomic"
    IMAGING = "imaging"
    PATIENT_IDS = "patient_ids"
    METADATA = "metadata"

    @classmethod
    def normalize(cls, file_type: str) -> str:
        return file_type.lower().strip()

    @classmethod
    def get_prefix_for_path(cls, filetype: str) -> str | None:
        if os.getenv("CONTEXT_MODE") == "TEST":
            return DOCKER_FOLDER_TEST
        else:
            if filetype.lower() in [Profile.PHENOTYPIC, Profile.CLINICAL, Profile.GENOMIC, Profile.IMAGING, Profile.MEDICINE, Profile.DIAGNOSIS]:
                return DOCKER_FOLDER_DATA
            elif filetype.lower() in [Profile.METADATA]:
                return DOCKER_FOLDER_METADATA
            # elif filetype.lower() == Profile.PATIENT_IDS:
            #     return DOCKER_FOLDER_ANONYMIZED_PATIENT_IDS
            else:
                return None

    @classmethod
    def get_preprocess_data_filename(cls, filetype) -> str:
        if filetype == Profile.PHENOTYPIC:
            return "phenotypic.csv"
        elif filetype == Profile.CLINICAL:
            return "samples.csv"
        elif filetype == Profile.DIAGNOSIS:
            return "diagnosis.csv"
        else:
            return None



from __future__ import annotations

import dataclasses
import json
import os
import re

import bson
import pymongo
from bson.json_util import loads
from filesplit.split import Split
from jsonlines import jsonlines
from pymongo import MongoClient
from pymongo.command_cursor import CommandCursor
from pymongo.cursor import Cursor

from constants.defaults import MAX_FILE_SIZE
from constants.methods import factory
from database.Execution import Execution
from database.Operators import Operators
from enums.TableNames import TableNames
from utils.file_utils import get_json_resource_file
from utils.setup_logger import log


@dataclasses.dataclass()
class Database:
    """
    The class Database represents the underlying MongoDB database: the connection, the database itself and
    auxiliary functions to make interactions with the database object (insert, select, ...).
    """

    SERVER_TIMEOUT = 5000
    execution: Execution
    # DO NOT DECLARE THOSE FIELDS HERE TO NOT ADD THEM TO ASDICT(),
    # because they are not thread-safe, thus are not pickable, thus cannot be jsonified
    # client: MongoClient = dataclasses.field(init=False, repr=False)
    # db: Any = dataclasses.field(init=False, repr=False)

    def __post_init__(self):
        """
        Initiate a new connection to a MongoDB client, reachable based on the given connection string, and initialize
        class members.
        """

        # 1. connect to the Mongo client
        try:
            # mongodb://localhost:27018/ -> goes through the host machine (open port - no need to have Mongo on the machine)
            # mongodb://mongo:27017/ -> mongo=Docker service name. goes through Docker network (no need to have Mongo on the machine)
            # Starting with version 3.0 the MongoClient constructor no longer blocks while connecting to the server or servers,
            # and it no longer raises pymongo (ConnectionFailure, ConfigurationError) errors.
            # Instead, the constructor returns immediately and launches the connection process on background threads.
            # You can check if the server is available with a ping.
            # w=0 disable acknowledgments from MongoDB (they are not necessary because we don't check them) to speedup write operations
            self.client = MongoClient(host=self.execution.db_connection, serverSelectionTimeoutMS=Database.SERVER_TIMEOUT, w=0)  # timeout after 5 sec instead of 20 (the default)
            log.info(type(self.client))
        except Exception:
            raise ConnectionError(f"Could not connect to the MongoDB client located at {self.execution.db_connection} and with a timeout of {Database.SERVER_TIMEOUT} ms.")

        # 2 check if the client is running well
        self.check_server_is_up()
        # if we reach this point, the MongoDB client runs correctly.
        log.info(f"The MongoDB client, located at {self.execution.db_connection}, could be accessed properly.")

        # 3. access the database
        log.info(f"drop db is: {self.execution.db_drop}")
        if self.execution.db_drop:
            self.drop_db()
            self.db = self.client[self.execution.db_name]
            log.info(type(self.db))
            self.drop_table(table_name=TableNames.STATS_DB)
            self.drop_table(table_name=TableNames.STATS_TIME)
            self.drop_table(table_name=TableNames.STATS_QUALITY)
        else:
            self.db = self.client[self.execution.db_name]
            log.info(type(self.db))

        log.debug(f"the connection string is: {self.execution.db_connection}")
        log.debug(f"the new MongoClient is: {self.client}")
        log.debug(f"the database is: {self.db}")

    def check_server_is_up(self) -> None:
        """
        Send a ping to confirm a successful connection.
        :return: A boolean being whether the MongoDB client is up.
        """
        try:
            self.client.admin.command('ping')
        except Exception:
            raise ConnectionError(f"The MongoDB client located at {self.execution.db_connection} could not be accessed properly.")

    def check_table_exists(self, table_name: str) -> bool:
        return table_name in self.db.list_collection_names()

    def drop_table(self, table_name: str) -> None:
        self.db.drop_collection(table_name)

    def drop_db(self) -> None:
        """
        Drop the current database.
        :return: Nothing.
        """
        if self.execution.db_drop:
            log.info(f"WARNING: The database {self.execution.db_name} will be dropped!", )
            self.client.drop_database(name_or_database=self.execution.db_name)

    def close(self) -> None:
        self.client.close()

    def insert_one_tuple(self, table_name: str, one_tuple: dict) -> None:
        # log.info(f"In table {table_name}, insert {one_tuple}")
        self.db[table_name].insert_one(one_tuple)

    def insert_many_tuples(self, table_name: str, tuples: list[dict] | tuple) -> None:
        """
        Insert the given tuples in the specified table.
        :param table_name: A string being the table name in which to insert the tuples.
        :param tuples: A list of dicts being the tuples to insert.
        """
        _ = self.db[table_name].insert_many(tuples, ordered=False)

    def update_one_tuple(self, table_name: str, filter_dict: dict, update: dict) -> None:
        _ = self.db[table_name].update_one(filter=filter_dict, update=self.create_update_stmt(the_tuple=update))

    def create_update_stmt(self, the_tuple: dict):
        # if self.execution.db_upsert_policy == UpsertPolicy.DO_NOTHING:
        #     # insert the document if it does not exist
        #     # otherwise, do nothing
        #     return {"$setOnInsert": the_tuple}
        # else:
        # insert the document if it does not exist
        # otherwise, replace it
        #     return {"$set": the_tuple}
        return {"$set": the_tuple}

    def upsert_one_tuple(self, table_name: str, unique_variables: list[str], one_tuple: dict) -> None:
        # filter_dict should only contain the fields on which we want a Resource to be unique,
        # e.g., name for Hospital instances, ID for Patient instances,
        #       the combination of Patient, Hospital, Clinical and LabFeature instances for LabRecord instances
        #       see https://github.com/Nelly-Barret/BETTER-fairificator/issues/3
        # one_tuple contains the Resource itself (with all its fields; as a JSON dict)
        # use $setOnInsert instead of $set to not modify the existing tuple if it already exists in the DB
        filter_dict = {}
        for unique_variable in unique_variables:
            try:
                filter_dict[unique_variable] = one_tuple[unique_variable]
            except Exception:
                raise KeyError(f"The tuple does not contains the attribute '{unique_variable}, thus the upsert cannot refer to it.")
        update_stmt = self.create_update_stmt(the_tuple=one_tuple)
        self.db[table_name].find_one_and_update(filter=filter_dict, update=update_stmt, upsert=True)

    def upsert_one_batch_of_tuples(self, table_name: str, unique_variables: list[str], the_batch: list[dict], ordered: bool) -> None:
        """

        :param unique_variables:
        :param table_name:
        :param the_batch:
        :return: The `filter_dict` to know exactly which fields have been used for the upsert (some of the given fields in unique_variables may not exist in some instances)
        """
        log.info(f"Upsert one batch of {len(the_batch)} tuples with unique variables being {unique_variables}")
        operations = [pymongo.UpdateOne(
            filter={unique_variable: one_tuple[unique_variable] for unique_variable in unique_variables if unique_variable in one_tuple},
            update=self.create_update_stmt(the_tuple=one_tuple), upsert=True)
            for one_tuple in the_batch]
        # July 18th, 2024: bulk_write modifies the hospital lists in Transform (even if I use deep copies everywhere)
        # It changes (only?) the timestamp value with +1/100, e.g., 2024-07-18T14:34:32Z becomes 2024-07-18T14:34:33Z
        # in the tests I use a delta to compare datetime
        self.db[table_name].bulk_write(operations, ordered=ordered)

    def retrieve_mapping(self, table_name: str, key_fields: str, value_fields: str, filter_dict: dict):
        # TODO Nelly: add a distinct to the find
        cursor = self.find_operation(table_name=table_name, filter_dict=filter_dict, projection={key_fields: 1, value_fields: 1})
        mapping = {}
        for result in cursor:
            projected_key = result
            for one_key in key_fields.split("."):
                # this covers the case when the key of the mapping is a nested field, e.g., identifier.value
                projected_key = projected_key[one_key]
            projected_value = result
            for one_value in value_fields.split("."):
                # this covers the case when the value of the mapping is a nested field, e.g., ontology_resource.label
                projected_value = projected_value[one_value]
            mapping[projected_key] = projected_value
        return mapping

    def load_json_in_table(self, table_name: str, unique_variables: list[str], dataset_id: int) -> None:
        self.load_json_in_table_general(table_name=table_name, unique_variables=unique_variables, dataset_id=dataset_id, ordered=False)

    def load_json_in_table_for_tests(self, unique_variables: list[str], dataset_id: int) -> None:
        self.load_json_in_table_general(table_name=TableNames.TEST, unique_variables=unique_variables, dataset_id=dataset_id, ordered=True)

    def load_json_in_table_general(self, table_name: str, unique_variables: list[str], dataset_id: int, ordered: bool) -> None:
        log.info(f"Write {table_name} data in table {table_name} with unique variables {unique_variables}")
        first_file = True
        expected_filename = get_json_resource_file(self.execution.working_dir_current, dataset_id, table_name)
        if os.path.exists(expected_filename):
            # this is a big file with all records for patients, or records, or features, or hospitals
            # we split it in smaller files of 15Mo (the MogoDB limit is 16Mo)
            # and send each chunk to the db
            split = Split(expected_filename, self.execution.working_dir_current)
            split.bysize(MAX_FILE_SIZE, newline=True)
            counter_files = 0
            regex_chunk_filename = re.compile(f"{dataset_id}{table_name}_[0-9]+\\.jsonl")
            total_count_files = sum([1 if regex_chunk_filename.search(elem) else 0 for elem in os.listdir(self.execution.working_dir_current)])
            for chunk_filename in os.listdir(self.execution.working_dir_current):
                if regex_chunk_filename.search(chunk_filename):
                    with jsonlines.open(os.path.join(self.execution.working_dir_current, chunk_filename), "r") as json_datafile:
                        if first_file:
                            # first, create an index on the unique variables to speed up the upsert (which checks whether each document already exists)
                            # we do this only if we have data for that kind of data
                            log.info(f"For table {table_name}, creating unique index {unique_variables}")
                            self.create_unique_index(table_name=table_name, columns={elem: 1 for elem in unique_variables})
                            first_file = False
                        # the chunk file is a JSON-by-line file, meaning that each record is on a line, with no separating comma and no encompassing array
                        # this needs to be added back to the JSON read string before parsing it
                        tuples = [obj for obj in json_datafile]  # this transforms the JSONL file to a list of objects
                        tuples = bson.json_util.loads(json.dumps(tuples))  # we need to read the objects with bson to interpret dates
                        self.upsert_one_batch_of_tuples(table_name=table_name, unique_variables=unique_variables, the_batch=tuples, ordered=ordered)
                        counter_files += 1
                        if counter_files % 5 == 0:
                            log.debug(f"Table {table_name}, loaded {counter_files}/{total_count_files}")
            log.debug(f"Table {table_name}, loaded {counter_files}/{total_count_files}")
        else:
            log.debug(f"Table {table_name}, no file to load.")

    def find_operation(self, table_name: str, filter_dict: dict, projection: dict) -> Cursor:
        """
        Perform a find operation (SELECT * FROM x WHERE filter_dict) in a given table.
        :param table_name: A string being the table name in which the find operation is performed.
        :param filter_dict: A dict being the set of filters (conditions) to apply on the data in the given table. Give {} to not apply any filter.
        :param projection: A dict being the set of projections (selections) to apply on the data in the given table. Give {} to return all fields.
        :return: A Cursor on the results, i.e., filtered data.
        """
        return self.db[table_name].find(filter_dict, projection)

    def find_distinct_operation(self, table_name: str, key: str, filter_dict: dict):
        """
        Perform a distinct operation on a field "key", with some filters on instances ("filter")
        :param table_name: A string being the table name in which the find operation is performed.
        :param key: A string being the key on which to apply the distinct; this is also the returned field.
        :param filter_dict: A dict being the set of filters (conditions) to apply on the data in the given table. Give {} to apply no filter.
        :return: A Cursor on the results, i.e., the distinct results.
        """
        return self.db[table_name].distinct(key, filter_dict)

    def count_documents(self, table_name: str, filter_dict: dict) -> int:
        """
        Count the number of documents in a table and matching a given filter.
        :param table_name: A string being the table name in which the count operation is performed.
        :param filter_dict: A dict being the set of filters to be applied on the documents.
        :return: An integer being the number of documents matched by the given filter.
        """
        return self.db[table_name].count_documents(filter_dict)

    def inverse_inner_join(self, name_table_1: str, name_table_2: str, foreign_field: str, local_field: str, lookup_name: str) -> CommandCursor:
        operations = [
            Operators.lookup(join_table_name=name_table_2, foreign_field=foreign_field, local_field=local_field, lookup_field_name=lookup_name, let=None, pipeline=None),
            Operators.match(field=lookup_name, value={"$eq": []}, is_regex=False),
            Operators.set_variables([{"name": "_id", "operation": 0}])
        ]
        return self.db[name_table_1].aggregate(operations)

    def list_existing_indexes(self, table_name: str) -> list:
        index_list = [res.values() for res in self.db[table_name].list_indexes()]
        log.info(index_list)
        return index_list

    def create_unique_index(self, table_name: str, columns: dict) -> None:
        """
        Create a unique constraint/index on a (set of) column(s).
        :param table_name: A string being the table name on which the index will be created.
        :param columns: A dict being the set of columns to be included in the index. It may contain only one entry if
        only one column should be unique. The parameter should be of the form { "colA": 1, ... }.
        :return: Nothing.
        """
        log.info(f"create unique index in {table_name} on columns {columns}")
        self.db[table_name].create_index(columns, unique=True)

    def create_non_unique_index(self, table_name: str, columns: dict) -> None:
        """
        Create an index on a (set of) column(s) for which uniqueness is not guaranteed.
        :param table_name: A string being the table name on which the index will be created.
        :param columns: A dict being the set of columns to be included in the index. It may contain only one entry if
        only one column should be unique. The parameter should be of the form { "colA": 1, ... }.
        :return: Nothing.
        """
        log.info(f"create non-unique index in {table_name} on columns {columns}")
        self.db[table_name].create_index(columns, unique=False)

    def create_on_demand_view(self, table_name: str, view_name: str, pipeline: list) -> None:
        self.drop_table(table_name=view_name)
        self.db.create_collection(view_name, viewOn=table_name, pipeline=pipeline)

    def refresh_on_demand_view(self, table_name: str, view_name: str, pipeline: list) -> None:
        # there is no explicit mechanism to refresh a materialized (on-demand) view
        # one need to recompute it
        self.create_on_demand_view(table_name=table_name, view_name=view_name, pipeline=pipeline)

    def get_min_or_max_value(self, table_name: str, field: str, sort_order: int) -> int | float:
        operations = []
        last_field = field.split(".")[-1]

        # if from_string:
        #     # we need to parse the string to long
        #     operations.append(Operators.project(field=field, projected_value={"split_var": {"$split": [f"${field}", DELIMITER_RESOURCE_ID]}}))
        #     operations.append(Operators.unwind(field="split_var"))
        #     operations.append(Operators.match(field="split_var", value="^[0-9]+$", is_regex=True))  # only numbers
        #     operations.append(Operators.group_by(group_key={"var": "$identifier"}, groups=[{"name": "min_max", "operator": "$max", "field": {"$toLong": "$split_var"}}]))
        #     operations.append(Operators.sort(field="min_max", sort_order=sort_order))
        #     operations.append(Operators.limit(1))
        #
        #     # better_default > db["ExaminationRecord"].aggregate([
        #     #     {"$project": {"identifier.value": {"$split": ["$identifier.value", "/"]}}},
        #     #     {"$unwind": "$identifier.value"},
        #     #     {"$match": {"identifier.value": / [0 - 9] + /}},
        #     # {"$group": {"_id": "identifier.value", "Max": {"$max": {"$toLong": "$identifier.value"}}}}
        #     # ])
        # else:
        if "." in field:
            # this field is a nested one, we only keep the deepest one,
            # e.g. for { "identifier": {"value": 1}} we keep { "value": 1}
            operations.append(Operators.project(field=field, projected_value=last_field))
            operations.append(Operators.sort(field=last_field, sort_order=sort_order))
            operations.append(Operators.limit(1))
        else:
            operations.append(Operators.project(field=field, projected_value=None))
            operations.append(Operators.sort(field=field, sort_order=sort_order))
            operations.append(Operators.limit(1))
        cursor = self.db[table_name].aggregate(operations)
        for result in cursor:
            # There should be only one result, so we can return directly the min or max value
            # if from_string:
            #     return result["min_max"]
            # else:
            if "." in field:
                return result[last_field]
            else:
                return result[field]
        return -1

    def get_max_value(self, table_name: str, field: str) -> int | float:
        return self.get_min_or_max_value(table_name=table_name, field=field, sort_order=-1)

    def get_min_value(self, table_name: str, field: str) -> int | float:
        return self.get_min_or_max_value(table_name=table_name, field=field, sort_order=1)

    def get_max_resource_counter_id(self) -> int:
        max_value = -1
        for table_name in TableNames.data_tables():
            current_max_identifier = self.get_max_value(table_name=table_name, field="identifier") # cannot use Resource.IDENTIFIER_ because ti leads to a circular dependency
            log.info(f"current max identifier for {table_name} is {current_max_identifier}")
            if current_max_identifier is not None:
                try:
                    current_max_identifier = int(current_max_identifier)
                    if current_max_identifier > max_value:
                        max_value = current_max_identifier
                except ValueError:
                    # this identifier is not an integer, e.g., a Clinical base ID like 24DL54
                    # we simply ignore it and try to find the next maximum integer ID
                    pass
            else:
                # the table is not created yet (this happens when we start from a fresh new DB, thus we skip this)
                pass
        return max_value

    def db_exists(self, db_name: str) -> bool:
        list_dbs = self.client.list_databases()
        for db in list_dbs:
            if db['name'] == db_name:
                return True
        return False

    def to_json(self):
        # for this class specifically, we cannot use the default factory
        # because PyMongo objects are not serializable
        return dataclasses.asdict(self, dict_factory=factory)
        # return {
        #     "execution": self.execution.to_json(),
        #     "mongo_client": str(self.client)
        # }

    def __str__(self):
        return json.dumps(self.to_json())



import dataclasses

from database.Database import Database
from entities.Record import Record
from entities.Resource import Resource
from enums.TableNames import TableNames
from main_statistics.MainStatistics import MainStatistics


@dataclasses.dataclass()
class DatabaseStatistics(MainStatistics):
    counts_instances: dict = dataclasses.field(default_factory=dict)
    records_with_no_value: dict = dataclasses.field(default_factory=dict)
    records_with_no_value_per_instantiate: dict = dataclasses.field(default_factory=dict)
    cc_with_no_text_per_table: dict = dataclasses.field(default_factory=dict)
    cc_with_no_onto_resource_per_table: dict = dataclasses.field(default_factory=dict)
    unknown_patient_refs_per_table: dict = dataclasses.field(default_factory=dict)
    unknown_hospital_refs_per_table: dict = dataclasses.field(default_factory=dict)
    unknown_feat_refs_in_records: dict = dataclasses.field(default_factory=dict)

    def compute_stats(self, database: Database):
        if self.record_stats:
            self.compute_counts_instances(database=database)
            self.compute_rec_with_no_value(database=database)
            self.compute_rec_with_no_value_per_instantiate(database=database)
            self.compute_onto_resources_with_no_label_per_table(database=database)
            self.compute_unknown_patient_refs_per_record_table(database=database)
            self.compute_unknown_hospital_refs_per_record_table(database=database)
            self.compute_unknown_feat_refs_in_records(database=database)

    @classmethod
    def jsonify_tuple(cls, one_tuple: dict) -> dict:
        return {key: str(value) for key, value in one_tuple.items()}

    def compute_counts_instances(self, database: Database) -> None:
        for table_name in [TableNames.HOSPITAL, TableNames.PATIENT, TableNames.FEATURE, TableNames.RECORD]:
            if table_name not in self.counts_instances:
                self.counts_instances[table_name] = {}
            self.counts_instances[table_name] = database.count_documents(table_name=table_name, filter_dict={})

    def compute_rec_with_no_value(self, database: Database) -> None:
        # for each RecordX, count the number of instances with no field "value"
        no_val_records = [DatabaseStatistics.jsonify_tuple(res) for res in database.find_operation(table_name=TableNames.RECORD, filter_dict={Record.VALUE_: {"$exists": 0}}, projection={"_id": 0})]
        self.records_with_no_value[TableNames.RECORD] = {"elements": no_val_records, "size": len(no_val_records)}

    def compute_rec_with_no_value_per_instantiate(self, database: Database) -> None:
        # for each RecordX, get the distinct list of "instantiates" references that do not have a value in the Record
        # db["LaboratoryRecord"].distinct("instantiates", {"value": {"$exists": 0}})
        # this query returns something like [ { reference: '83' }, { reference: '87' } ]
        # then we process it to return a dict <ref. id, count>, e.g. { "83": {"elements": [...], "size": 5}, "87": {...} }
        instantiates_no_value = [res for res in database.find_distinct_operation(table_name=TableNames.RECORD, key=Record.INSTANTIATES_, filter_dict={Record.VALUE_: {"$exists": 0}})]
        records_with_no_val_per_instantiate = {}
        for instantiate_ref in instantiates_no_value:
            records_with_no_val_per_instantiate[instantiate_ref] = [DatabaseStatistics.jsonify_tuple(res) for res in database.find_operation(table_name=TableNames.RECORD, filter_dict={Record.INSTANTIATES_: instantiate_ref, Record.VALUE_: {"$exists": 0}}, projection={"_id": 0})]
            if TableNames.RECORD not in self.records_with_no_value_per_instantiate:
                self.records_with_no_value_per_instantiate[TableNames.RECORD] = {}
            self.records_with_no_value_per_instantiate[TableNames.RECORD][instantiate_ref] = {"elements": records_with_no_val_per_instantiate, "size": len(records_with_no_val_per_instantiate)}

    def compute_onto_resources_with_no_label_per_table(self, database: Database) -> None:
        # db["LaboratoryFeature"].find({ "ontology_resource.label": "" })
        for table_name in [TableNames.FEATURE, TableNames.RECORD]:
            if table_name not in self.cc_with_no_text_per_table:
                self.cc_with_no_text_per_table[table_name] = {}
            no_text_cc = [DatabaseStatistics.jsonify_tuple(res) for res in database.find_operation(table_name=table_name, filter_dict={"ontology_resource.label": ""}, projection={"_id": 0})]
            self.cc_with_no_text_per_table[table_name] = {"elements": no_text_cc, "size": len(no_text_cc)}

    def compute_unknown_patient_refs_per_record_table(self, database: Database) -> None:
        unknown_patient_refs = [DatabaseStatistics.jsonify_tuple(res) for res in database.inverse_inner_join(name_table_1=TableNames.RECORD, name_table_2=TableNames.PATIENT, foreign_field=Resource.IDENTIFIER_, local_field=Record.SUBJECT_, lookup_name="KnownRefs")]
        self.unknown_patient_refs_per_table[TableNames.RECORD] = {"elements": unknown_patient_refs, "size": len(unknown_patient_refs)}

    def compute_unknown_hospital_refs_per_record_table(self, database: Database) -> None:
        unknown_hospital_refs = [DatabaseStatistics.jsonify_tuple(res) for res in database.inverse_inner_join(name_table_1=TableNames.RECORD, name_table_2=TableNames.HOSPITAL, foreign_field=Resource.IDENTIFIER_, local_field=Record.REG_BY_, lookup_name="KnownRefs")]
        self.unknown_hospital_refs_per_table[TableNames.RECORD] = {"elements": unknown_hospital_refs, "size": len(unknown_hospital_refs)}

    def compute_unknown_feat_refs_in_records(self, database: Database) -> None:
        unknown_refs = [DatabaseStatistics.jsonify_tuple(res) for res in database.inverse_inner_join(name_table_1=TableNames.RECORD, name_table_2=TableNames.FEATURE, foreign_field=Resource.IDENTIFIER_, local_field=Record.INSTANTIATES_, lookup_name="KnownRefs")]
        self.unknown_feat_refs_in_records = {"elements": unknown_refs, "size": len(unknown_refs)}



import dataclasses

from entities.Record import Record
from enums.Profile import Profile
from enums.TableNames import TableNames


@dataclasses.dataclass()
class GenomicRecord(Record):
    vcf: str
    entity_type: str = f"{Profile.GENOMIC}{TableNames.RECORD}"



from typing import Any

import numpy as np
import pandas as pd

from enums.EnumAsClass import EnumAsClass
from utils.setup_logger import log

from utils.str_utils import process_spaces


class Ontologies(EnumAsClass):
    # ontology names HAVE TO be normalized by hand here because we can't refer to static methods
    # because they do not exist yet in the execution context
    SNOMEDCT = {"name": "snomedct", "url": "http://snomed.info/sct"}
    LOINC = {"name": "loinc", "url": "http://loinc.org"}
    CLIR = {"name": "clir", "url": "https://clir.mayo.edu"}
    PUBCHEM = {"name": "pubchem", "url": "https://pubchem.ncbi.nlm.nih.gov"}
    GSSO = {"name": "gsso", "url": "http://purl.obolibrary.org/obo"}
    ORPHANET = {"name": "orpha", "url": "https://www.orpha.net/"}
    GENE_ONTOLOGY = {"name": "geneontology", "url": "https://amigo.geneontology.org/amigo"}
    OMIM = {"name": "omim", "url": "https://omim.org"}
    HGNC = {"name": "hgnc", "url": "https://rest.ensembl.org/"}
    NONE = {"name": "none", "url": "none"}
    HPO = {"name": "hp", "url": "https://clinicaltables.nlm.nih.gov/api/hpo/"}

    @classmethod
    def get_enum_from_name(cls, ontology_name: str) -> dict:
        if ontology_name == "":
            return {}
        else:
            for existing_ontology in Ontologies.values():
                if existing_ontology["name"] == ontology_name:
                    return existing_ontology  # return the ontology enum
            return {}

    @classmethod
    def get_enum_from_url(cls, ontology_url: str):
        for existing_ontology in Ontologies.values():
            if existing_ontology["url"] == ontology_url:
                return existing_ontology  # return the ontology enum
        return ""

    @classmethod
    def get_names(cls):
        return [ontology["name"] for ontology in Ontologies.values()]

    @classmethod
    def get_urls(cls):
        return [ontology["url"] for ontology in Ontologies.values()]

    @classmethod
    def normalize_name(cls, ontology_name: str) -> str:
        if ontology_name == "":
            return ""
        else:
            return process_spaces(input_string=ontology_name).lower().replace(" ", "").replace("_", "")

    @classmethod
    def normalize_code(cls, ontology_code: str) -> str:
        if ontology_code == "":
            return ""
        else:
            return process_spaces(input_string=ontology_code).lower().replace(" ", "")



from enums.EnumAsClass import EnumAsClass


class UserAgents(EnumAsClass):
    MOZILLA_WIN_109 = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36'
    MOZILLA_WIN_108 = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36'
    MOZILLA_LIN_108 = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36'
    MOZILLA_MAC_109 = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36'
    MOZILLA_MAC_108 = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36'
    MOZILLA_MAC_16 = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15'
    MOZILLA_MAC_13 = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 13_1) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15'



import pandas as pd
from pandas import DataFrame

from database.Execution import Execution
from enums.Profile import Profile
from preprocessing.Preprocess import Preprocess


class PreprocessCovid(Preprocess):
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str):
        super().__init__(execution=execution, data=data, metadata=metadata, profile=profile)

    def preprocess(self):
        if self.profile == Profile.CLINICAL:
            # process samples data to transpose them
            self.data.drop(["hospital", "interpolated", "time_start", "time_end"], axis=1, inplace=True)
            new_df_as_json = {}
            columns = list(self.data["test"].unique())
            all_ids = list(self.data["id"].unique())
            one_patient = {}
            all_patients = []

            if "id" not in new_df_as_json:
                new_df_as_json["id"] = []
            for column in columns:
                new_df_as_json[column] = []

            for pid in all_ids:
                # print(f"****{pid}****")
                tests_for_patient = DataFrame(self.data.loc[self.data["id"] == pid]["test"])
                tests_for_patient = tests_for_patient.reset_index(drop=True)
                values_for_patient = DataFrame(self.data.loc[self.data["id"] == pid]["value"])
                values_for_patient = values_for_patient.reset_index(drop=True)
                max_occurrence = max(tests_for_patient.value_counts())

                one_patient["id"] = [pid for _ in range(max_occurrence)]
                for column in columns:
                    one_patient[column] = [None for _ in range(max_occurrence)]
                # print(f"one_patient = {one_patient}")
                # print(f"tests_for_patient = {tests_for_patient}")
                # print(f"values_for_patient = {values_for_patient}")
                for i in range(len(values_for_patient)):
                    the_test = tests_for_patient.iloc[i].iloc[0]
                    the_value = values_for_patient.iloc[i].iloc[0]
                    # print(f"value={the_value}; ")
                    # print(f"tests_for_patient[i]={the_test}")
                    # print(f"index {i%max_occurrence} in array: {one_patient[the_test]}")
                    one_patient[the_test][i%max_occurrence] = the_value
                    # print(f"one_patient = {one_patient}")
                all_patients.append(one_patient)
                one_patient = {}

            final_json = {}
            for patient_as_json in all_patients:
                for key in patient_as_json:
                    if key not in final_json:
                        final_json[key] = []
                    final_json[key].extend(patient_as_json[key])
            # print(json.dumps(all_patients))
            # print(json.dumps(final_json))

            self.data = pd.DataFrame(final_json)
            self.data["sid"] = [f"s{i}" for i in range(1, len(self.data)+1)]



import dataclasses

from entities.Resource import Resource
from enums.TableNames import TableNames


@dataclasses.dataclass()
class Hospital(Resource):
    name: str
    entity_type: str = f"{TableNames.HOSPITAL}"

    # keys to be used when writing JSON or queries
    # those names have to exactly match the variables names declared in entity classes
    NAME_ = "name"

    # keys to be used when using an entity attribute as a query variable
    NAME__ = f"${NAME_}"



import itertools
import json
import os

import pandas as pd
from pandas import DataFrame

from database.Database import Database
from database.Execution import Execution
from entities.Feature import Feature
from entities.OntologyResource import OntologyResource
from enums.DataTypes import DataTypes
from enums.HospitalNames import HospitalNames
from enums.MetadataColumns import MetadataColumns
from enums.Ontologies import Ontologies
from enums.Profile import Profile
from enums.TableNames import TableNames
from enums.VcfColumns import vcf_columns
from enums.Visibility import Visibility
from etl.Task import Task
from preprocessing.PreprocessingTask import PreprocessingTask
from main_statistics.QualityStatistics import QualityStatistics
from utils.file_utils import read_tabular_file_as_string
from utils.setup_logger import log


class Extract(Task):

    def __init__(self, metadata: DataFrame, profile: str, database: Database, execution: Execution, quality_stats: QualityStatistics):
        super().__init__(database=database, execution=execution, quality_stats=quality_stats)
        self.data = None
        self.metadata = metadata
        self.profile = Profile.normalize(profile)
        self.columns_dataset_all_profiles = None
        # self.mapping_categorical_value_to_onto_resource = {}  # <categorical value label ("JSON_values" column), OntologyResource>
        self.mapping_column_to_categorical_value = {}  # <column name, list of normalized accepted values>
        self.mapping_column_to_vartype = {}  # <column name, var type ("vartype" column)>
        self.mapping_column_to_unit = {}  # <column name, unit provided in the metadata>
        self.mapping_column_to_domain = {}  # <column name, <min: x, max: y> or <accepted_values: [...]>>

    def run(self) -> None:

        # filter and normalize metadata
        # filter: keep metadata about the current triplet <dataset, profile, hospital>
        # normalize: the header and the values
        self.filter_metadata_file()
        self.normalize_metadata_file()

        # filter and normalize data
        # filter: remove data columns that are not described in the metadata
        # normalize: the header
        if self.metadata is not None:
            # preprocess input to have all the necessary data, as described in the metadata
            self.load_tabular_data()
            self.pre_process_data_file()
            self.filter_data_file()
            self.normalize_data_file()

            # compute mappings (categories, units and domains)
            self.compute_mapping_categorical_value_to_onto_resource()
            self.compute_column_to_unit()
            self.compute_column_to_domain()

            # finally, export both metadata and data
            # we need to wait because metadata can still be updated by the data pre-processing task
            self.export_metadata_to_csv_for_generative_ai()
            self.export_data_to_csv_for_generative_ai()

    def export_metadata_to_csv_for_generative_ai(self):
        exported_filepath = os.path.join(self.execution.working_dir_current, "exported_metadata.csv")
        if os.path.exists(exported_filepath) and os.path.getsize(exported_filepath) > 0:
            # existing metadata, we happen
            self.metadata.to_csv(exported_filepath, mode="a", header=False, index=False)
        else:
            self.metadata.to_csv(exported_filepath, index=False)

    def export_data_to_csv_for_generative_ai(self):
        filename = os.path.basename(self.execution.current_filepath)  # it contains the .csv
        filename = filename[0:filename.index(".")]
        exported_filepath = os.path.join(self.execution.working_dir_current, f"exported_{filename}.csv")
        log.info(exported_filepath)
        if os.path.exists(exported_filepath) and os.path.getsize(exported_filepath) > 0:
            # existing data, we happen the new columns (thus we need to read data as a df because "a" only appends rows)
            existing_data = pd.read_csv(exported_filepath, index_col=False)
            self.data.columns = self.data.columns.map(str)
            existing_data = existing_data.merge(self.data, on=self.execution.patient_id_column_name, how="outer")  # append current data as columns next to existing data
            existing_data.to_csv(exported_filepath, index=False)
        else:
            self.data.columns = self.data.columns.map(str)  # keep column names as strings (to avoid convert HPO terms as integers, which removes leading 000)
            self.data.to_csv(exported_filepath, index=False)

    def filter_metadata_file(self) -> None:
        # Normalize the header, e.g., "Significato it" becomes "significato_it"
        # this also normalizes hospital names if they are in the header (UC 2 and UC 3)
        self.metadata.rename(columns=lambda x: MetadataColumns.normalize_name(column_name=x), inplace=True)

        # normalize the profiles before filtering
        self.columns_dataset_all_profiles = self.metadata[MetadataColumns.COLUMN_NAME]  # this contains all feature of the dataset (no matter the profile)
        self.metadata.loc[:, MetadataColumns.PROFILE] = self.metadata[MetadataColumns.PROFILE].apply(lambda x: Profile.normalize(file_type=x))
        self.metadata = self.metadata[self.metadata[MetadataColumns.PROFILE].values == self.profile]

        # if the filtered metadata (by dataset and profile) is not empty, we check whether we need to further filter
        # metadata by hospital name
        if len(self.metadata) > 0:
            normalized_hospital_name = HospitalNames.normalize(self.execution.hospital_name)
            if normalized_hospital_name in self.metadata.columns:
                # the current hospital is in the metadata, we need to select the metadata line for which the
                # hospital column has 1
                self.metadata.loc[:, normalized_hospital_name] = self.metadata[normalized_hospital_name].apply(lambda x: MetadataColumns.normalize_value(column_value=x))
                self.metadata = self.metadata[self.metadata[normalized_hospital_name].isin([1, "1"])]
            else:
                # we have no column specifying a hospital name, so the metadata is only for the current hospital
                # thus, nothing to do
                pass

            # 6. reindex the remaining metadata because when dropping rows/columns, they keep their original index
            self.metadata.reset_index(drop=True, inplace=True)

            log.info(f"{len(self.metadata.columns)} columns and {len(self.metadata)} lines in the metadata file.")
        else:
            # no metadata for this combination of <dataset, profile>, so we skip it
            self.metadata = None
            log.info("metadata is None")

    def normalize_metadata_file(self) -> None:
        if self.metadata is not None:
            # Normalize ontology names (but not codes because they will be normalized within OntologyResource)
            self.metadata.loc[:, MetadataColumns.ONTO_NAME] = self.metadata[MetadataColumns.ONTO_NAME].apply(lambda value: Ontologies.normalize_name(ontology_name=value))

            # Normalize ETL type
            self.metadata.loc[:, MetadataColumns.ETL_TYPE] = self.metadata[MetadataColumns.ETL_TYPE].apply(lambda x: DataTypes.normalize(data_type=x))

            # Normalize visibility
            self.metadata.loc[:, MetadataColumns.VISIBILITY] = self.metadata[MetadataColumns.VISIBILITY].apply(lambda x: Visibility.normalize(visibility=x))

            # Normalize column name
            self.metadata.loc[:, MetadataColumns.COLUMN_NAME] = self.metadata[MetadataColumns.COLUMN_NAME].apply(lambda x: MetadataColumns.normalize_name(column_name=x))

            # Compute some stats about the metadata
            for row in self.metadata.itertuples(index=False):
                column_name = row[self.metadata.columns.get_loc(MetadataColumns.COLUMN_NAME)]
                etl_type = row[self.metadata.columns.get_loc(MetadataColumns.ETL_TYPE)]
                onto_name = row[self.metadata.columns.get_loc(MetadataColumns.ONTO_NAME)]
                onto_code = row[self.metadata.columns.get_loc(MetadataColumns.ONTO_CODE)]
                if onto_code == "":
                    self.quality_stats.add_column_with_no_ontology(column_name=column_name)
                if etl_type == "":
                    self.quality_stats.add_column_with_no_etl_type(column_name=column_name)
                else:
                    if etl_type not in DataTypes.values():
                        self.quality_stats.add_column_unknown_etl_type(column_name=column_name, etl_type=etl_type)
                if onto_name != "" and len(Ontologies.get_enum_from_name(onto_name)) == 0:
                    self.quality_stats.add_column_unknown_ontology(column_name=column_name, ontology_name=onto_name)
        else:
            # metadata is None because nothing remains from the filtering
            # skip the normalization (nothing to normalize)
            pass

    def load_tabular_data(self) -> None:
        log.info(f"Data filepath is {self.execution.current_filepath}")
        assert os.path.exists(self.execution.current_filepath), "The provided data file could not be found."
        self.data = read_tabular_file_as_string(filepath=self.execution.current_filepath)

    def normalize_data_file(self):
        # Normalize the data values
        # they will be cast to the right type (int, float, datetime) in the Transform step
        # issue 113: we do not normalize identifiers assigned by hospitals to avoid discrepancies (not the VCF filenames)
        columns_no_normalization = []
        columns_no_normalization.append(self.execution.patient_id_column_name)
        if self.execution.hospital_name in vcf_columns:
            columns_no_normalization.append(vcf_columns[self.execution.hospital_name])
        if self.execution.sample_id_column_name != "":
            columns_no_normalization.append(self.execution.sample_id_column_name)

        for column in self.data:
            if column not in columns_no_normalization:
                self.data.loc[:, column] = self.data[column].apply(lambda x: MetadataColumns.normalize_value(column_value=x))

        log.info(f"{len(self.data.columns)} columns and {len(self.data)} lines in the data file.")

    def pre_process_data_file(self) -> None:
        # preprocess data files, i.e., change the data DataFrame to fit the metadata
        # we do not write the pre-processed data to any new ile, we simply run the ETL with it
        # this avoids to (a) overwrite given data files and (b) to have filenames which differ from the metadata
        preprocessing_task = PreprocessingTask(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile, all_columns_dataset=self.columns_dataset_all_profiles)
        preprocessing_task.run()
        self.data = preprocessing_task.data
        self.metadata = preprocessing_task.metadata
        self.columns_dataset_all_profiles = preprocessing_task.all_columns_dataset  # update the list of features if some have been added during the data pre-processing

        # normalize column names
        self.data = self.data.rename(columns=lambda x: MetadataColumns.normalize_name(column_name=x))

    def filter_data_file(self) -> None:
        # Normalize column names ("sex", "dateOfBirth", "Ethnicity", etc.) to match column names described in the metadata
        self.data = self.data.rename(columns=lambda x: MetadataColumns.normalize_name(column_name=x))

        # removes the data columns that are NOT described in the metadata or that are explicitly marked as not to be loaded (except if this is an ID column)
        # if a column is described in the metadata but is not present in the data or this column is empty we keep it
        # because people took the time to describe it.
        # we record this column in the stats only if it is not described at all in the current file metadata
        # this is because we iteratively look at the metadata of each pair <dataset, profile>
        # and we do not want to record a column as "not described" if it is later described in another profile
        data_columns = list(set(self.data.columns))  # get the distinct list of columns
        expected_columns_dataset = list(self.columns_dataset_all_profiles.apply(lambda x: MetadataColumns.normalize_name(x)))  # https://git.rwth-aachen.de/padme-development/external/better/data-cataloging/etl/-/issues/282
        columns_to_drop = [data_column for data_column in data_columns if data_column not in expected_columns_dataset or (data_column in self.execution.columns_to_remove and data_column not in [self.execution.patient_id_column_name, self.execution.sample_id_column_name])]
        log.info(f"for profile {self.profile}, columns to drop are: {columns_to_drop}")
        self.data = self.data.drop(columns_to_drop, axis=1)  # axis=1 -> columns
        for column in columns_to_drop:
            log.info(f"Data column {column} is not described in the metadata, skipping it.")
            self.quality_stats.add_column_not_described_in_metadata(data_column_name=column)

        # for now, we only dropped columns that were not expected (because not described in the metadata)
        # now, we still need to filter the data based on the profile
        log.info(self.data.columns)
        # the metadata has already been filtered and has also been pre-processed (e.g., to add new metadata for the current profile)
        # be careful: the metadata may contain columns that are not present in the data - filtering with self.metadata[MetadataColumns.COLUMN_NAME].to_list() will produce an "unknown column" exception
        columns_to_keep = list(set(self.metadata[MetadataColumns.COLUMN_NAME].to_list()) & set(self.data.columns))
        log.info(columns_to_keep)
        self.data = self.data[columns_to_keep]

    def compute_mapping_categorical_value_to_onto_resource(self) -> None:
        # Apr 15, 2025: I cannot use the mapping category/OR
        # because in IMGGE all categorical values are encoded with numbers (1, 2, 3, etc.)
        # and thus this mapping lacks a level of nesting with the feature name to avoid writing agin and again new OR for categorical values
        # self.mapping_categorical_value_to_onto_resource = {}
        self.mapping_column_to_categorical_value = {}
        # 1. first, we retrieve the existing categorical values already transformed as OntologyResource
        # this will avoid to send again API queries to re-build already-built OntologyResource,
        # e.g., when starting from an existing DB (drop=False)
        existing_categorical_codeable_concepts = {}
        # the set of categorical values are defined in Features only, thus we can restrict the find to only those:
        # categorical_values_for_table_name = {'_id': ObjectId('...'), 'categorical_values': [{...}, {...}, ...]}
        categorical_values_for_table_name = self.database.find_operation(table_name=TableNames.FEATURE, filter_dict={Feature.CATEGORIES_: {"$exists": 1}}, projection={Feature.CATEGORIES_: 1})
        for one_tuple in categorical_values_for_table_name:
            # existing_categorical_value_for_table_name = [{...}, {...}, ...]}
            existing_categorical_values_for_table_name = one_tuple[Feature.CATEGORIES_]
            for encoded_categorical_value in existing_categorical_values_for_table_name:
                existing_or = OntologyResource.from_json(encoded_categorical_value, quality_stats=self.quality_stats)
                existing_categorical_codeable_concepts[existing_or.label] = existing_or

        # 2. then, we associate each column to its set of categorical values
        # if we already compute the cc of that value (e.g., several column have categorical values Yes/No/NA),
        # we do not recompute it and take it from the mapping
        for row in self.metadata.itertuples(index=False):
            column_name = row[self.metadata.columns.get_loc(MetadataColumns.COLUMN_NAME)]
            candidate_json_values = row[self.metadata.columns.get_loc(MetadataColumns.JSON_VALUES)]
            column_type = row[self.metadata.columns.get_loc(MetadataColumns.ETL_TYPE)]
            if candidate_json_values != "":
                # we get the possible categorical values for the column, e.g., F, or M, or NA for sex
                try:
                    json_categorical_values = json.loads(candidate_json_values)
                except Exception:
                    self.quality_stats.add_categorical_colum_with_unparseable_json(column_name=column_name, broken_json=candidate_json_values)
                    json_categorical_values = {}
                categories_for_column = {}  # we append this dictionary to the list only if it has at least one categorical value
                # { 'sex': {
                #   'm': {"system": "snomed", "code": "248153007", "label": "m (Male)"},
                #   'f': {"system": "snomed", "code": "248152002", "label": "f (Female)"},
                #   'inadeguata': {"system": "snomed", "code": "71978007", "label": "inadeguata (inadequate)"},
                #   ...
                # }, ...}
                for json_categorical_value in json_categorical_values:
                    normalized_categorical_value = MetadataColumns.normalize_value(json_categorical_value["value"])
                    if normalized_categorical_value not in categories_for_column:
                        # the categorical value does not exist yet in the mapping, thus:
                        # - it may be retrieved from the db and be added to the mapping
                        # - or, it may be computed for the first time
                        if normalized_categorical_value not in existing_categorical_codeable_concepts.keys():
                            # this is a categorical value that we have never seen (not even in previous executions),
                            # we need to create an OntologyResource for it from scratch
                            # json_categorical_value is of the form: {'value': 'X', 'explanation': 'some definition', 'onto_system_Y': 'onto_code_Z' }
                            if len(json_categorical_value) == 2 and "value" in json_categorical_value and "explanation" in json_categorical_value:
                                # this is the specific case when categories are not mapped to a code, but they are encoded
                                # therefore the mapping is simply the encoded value (1, 2, 3, etc) to the label (found in the "explanation" field)
                                # we need to use empty system and code instead of None otherwise NoneType exception
                                onto_resource = OntologyResource(system="", code="", label=json_categorical_value["explanation"], quality_stats=self.quality_stats)
                                categories_for_column[normalized_categorical_value] = onto_resource.to_json()
                            else:
                                for key, val in json_categorical_value.items():
                                    # for any key value pair that is not about the value or the explanation
                                    # (i.e., loinc and snomedct columns), we create an OntologyResource, which we add to the CodeableConcept
                                    # we need to do a loop because there may be several ontology terms for a single mapping
                                    if key != "value" and key != "explanation":
                                        # here, we do normalize the ontology name to be able to get the corresponding enum
                                        # however, we do not normalize the code, because it needs extra attention (due to spaces in post-coordinated codes, etc)
                                        ontology = Ontologies.get_enum_from_name(ontology_name=Ontologies.normalize_name(key))
                                        onto_resource = OntologyResource(system=ontology, code=val, label=None, quality_stats=self.quality_stats)
                                        if onto_resource.system != "" and onto_resource.code != "":
                                            categories_for_column[normalized_categorical_value] = onto_resource.to_json()
                                        else:
                                            # the ontology system is unknown or no code has been provided,
                                            # thus the OntologyResource contains only None fields,
                                            # thus we do not record it as a possible category
                                            pass

                        else:
                            # an OntologyResource already exists for this value (it has been retrieved from the db),
                            # we simply add it to the mapping
                            log.debug(f"The categorical value {normalized_categorical_value} already exists in the database as a CC. Taking it from here.")
                            categories_for_column[normalized_categorical_value] = existing_categorical_codeable_concepts[normalized_categorical_value].to_json()
                    if column_name not in self.mapping_column_to_categorical_value and len(categories_for_column) > 0:
                        # there is at least one category for this column
                        self.mapping_column_to_categorical_value[column_name] = categories_for_column
            else:
                # if this was supposed to be categorical (thus having values), we count it in the reporting
                # otherwise, this is not a categorical column (thus, everything is fine)
                if column_type == DataTypes.CATEGORY:
                    # log.info(f"no JSON categories for column {column_name}")
                    self.quality_stats.add_categorical_column_with_no_json(column_name=column_name)
        log.debug(f"{self.mapping_column_to_categorical_value}")

    def compute_column_to_unit(self) -> None:
        self.mapping_column_to_unit = {}

        for row in self.metadata.itertuples(index=False):
            unit = row[self.metadata.columns.get_loc(MetadataColumns.VAR_UNIT)]
            if unit != "":
                self.mapping_column_to_unit[row[self.metadata.columns.get_loc(MetadataColumns.COLUMN_NAME)]] = unit
            else:
                self.mapping_column_to_unit[row[self.metadata.columns.get_loc(MetadataColumns.COLUMN_NAME)]] = None

        if len(self.mapping_column_to_unit) > 10:
            log.debug(dict(itertools.islice(self.mapping_column_to_unit.items(), 10)))
        else:
            log.debug(self.mapping_column_to_unit)

    def compute_column_to_domain(self) -> None:
        self.mapping_column_to_domain = {}

        for row in self.metadata.itertuples(index=False):
            domain = row[self.metadata.columns.get_loc(MetadataColumns.DOMAIN)]
            try:
                domain = json.loads(domain)
                self.mapping_column_to_domain[row[self.metadata.columns.get_loc(MetadataColumns.COLUMN_NAME)]] = domain
            except:
                self.mapping_column_to_domain[row[self.metadata.columns.get_loc(MetadataColumns.COLUMN_NAME)]] = None
        log.debug(self.mapping_column_to_domain)



import dataclasses
from typing import Any

from entities.Resource import Resource


@dataclasses.dataclass()
class Record(Resource):
    has_subject: int
    registered_by: int
    instantiates: int
    value: Any
    dataset: str

    # keys to be used when writing JSON or queries
    # those names have to exactly match the variables names declared in entity classes
    SUBJECT_ = "has_subject"
    REG_BY_ = "registered_by"
    INSTANTIATES_ = "instantiates"
    VALUE_ = "value"
    BASE_ID_ = "base_id"

    # keys to be used when using an entity attribute as a query variable
    SUBJECT__ = f"${SUBJECT_}"
    REG_BY__ = f"${REG_BY_}"
    INSTANTIATES__ = f"${INSTANTIATES_}"
    VALUE__ = f"${VALUE_}"
    BASE_ID__ = f"${BASE_ID_}"

    # keys to be used for group by in queries
    INSTANTIATES___ = f"$_id.{INSTANTIATES_}"
    VALUE___ = f"$_id.{VALUE_}"



from typing import Any

import inflection

from constants.defaults import NAN_VALUES, DEFAULT_NAN_VALUE
from enums.EnumAsClass import EnumAsClass
from utils.str_utils import process_spaces


class MetadataColumns(EnumAsClass):
    # ontology names HAVE TO be normalized by hand here because we can't refer to static methods
    # because they do not exist yet in the execution context
    ONTO_NAME = "ontology"
    ONTO_CODE = "ontology_code"
    DATASET_NAME = "dataset"
    PROFILE = "profile"
    COLUMN_NAME = "name"
    SIGNIFICATION_EN = "description"
    VISIBILITY = "visibility"
    ETL_TYPE = "etl_type"
    VAR_UNIT = "dimension"
    DOMAIN = "domain"
    JSON_VALUES = "json_values"

    @classmethod
    def normalize_name(cls, column_name: str) -> str:
        column_name = process_spaces(input_string=column_name)
        return inflection.underscore(column_name).replace(" ", "_").lower()

    @classmethod
    def normalize_value(cls, column_value: str) -> Any:
        if column_value == "":
            # the cell was empty and kept as is during the data reading
            return ""
        else:
            normalized_value = process_spaces(input_string=str(column_value)).lower()
            if normalized_value in NAN_VALUES:
                # explicit NaN cell value
                return DEFAULT_NAN_VALUE
            else:
                # default case
                return normalized_value



import inflection

from enums.EnumAsClass import EnumAsClass
from utils.str_utils import process_spaces


class HospitalNames(EnumAsClass):
    # hospital names HAVE TO be normalized by hand here because we can't refer to static methods
    # because they do not exist yet in the execution context
    IT_BUZZI_UC1 = "it_buzzi_uc1"
    RS_IMGGE = "rs_imgge"
    ES_HSJD = "es_hsjd"
    IT_BUZZI_UC3 = "it_buzzi_uc3"
    ES_TERRASSA = "es_terrassa"
    DE_UKK = "de_ukk"
    ES_LAFE = "es_lafe"
    IL_HMC = "il_hmc"
    TEST_H1 = "test_h1"
    TEST_H2 = "test_h2"
    TEST_H3 = "test_h3"
    EXPES_EDA = "eda"
    EXPES_COVID = "covid"
    EXPES_KIDNEY = "kidney"

    @classmethod
    def normalize(cls, hospital_name: str) -> "":
        if hospital_name is not None:
            hospital_name = process_spaces(hospital_name)
            hospital_name = hospital_name.replace(" ", "_")  # add missing _ in hospital name if needed
            return inflection.underscore(hospital_name).lower()
        else:
            return ""



from enums.EnumAsClass import EnumAsClass


class TheTestFiles(EnumAsClass):
    # for TEST files only, we set the absolute path with the Docker path
    # this is because tests are to be run in Docker only and only have to share the folder datasets/test

    # original files for the Extract step
    ORIG_METADATA_PATH = "orig-metadata.csv"  # this describes several hospitals, several datasets per hospital, etc
    ORIG_PHENOTYPIC_PATH = "orig-data-phen.csv"
    ORIG_CLINICAL_PATH = "orig-data-clin.csv"
    ORIG_DIAGNOSIS_PATH = "orig-data-diag.csv"
    ORIG_DYNAMIC_PATH = "orig-data-dyn.csv"
    ORIG_GENOMICS_PATH = "orig-data-gen.csv"
    # ORIG_EMPTY_PIDS_PATH = "orig-empty-pids.json"
    # ORIG_FILLED_PIDS_PATH = "orig-filled-pids.json"

    # files obtained after the Extract step
    # ued for the Transform step
    EXTR_METADATA_PHENOTYPIC_PATH = "extr-metadata-phen.csv"
    EXTR_METADATA_CLINICAL_PATH = "extr-metadata-clin.csv"
    EXTR_METADATA_DIAGNOSIS_PATH = "extr-metadata-diag.csv"
    EXTR_METADATA_DYNAMIC_PATH = "extr-metadata-dyn.csv"
    EXTR_METADATA_GENOMICS_PATH = "extr-metadata-gen.csv"
    EXTR_PHENOTYPIC_DATA_PATH = "extr-data-phen.csv"
    EXTR_CLINICAL_DATA_PATH = "extr-data-clin.csv"
    EXTR_DIAGNOSIS_DATA_PATH = "extr-data-diag.csv"
    EXTR_DYNAMIC_DATA_PATH = "extr-data-dyn.csv"
    EXTR_GENOMICS_DATA_PATH = "extr-data-gen.csv"
    EXTR_PHENOTYPIC_COL_CAT_PATH = "extr-data-phen-column-categorical.json"
    EXTR_CLINICAL_COL_CAT_PATH = "extr-data-clin-column-categorical.json"
    EXTR_PHENOTYPIC_UNITS_PATH = "extr-data-phen-column-to-unit.json"
    EXTR_CLINICAL_TYPE_PATH = "extr-data-clin-column-to-type.json"
    EXTR_PHENOTYPIC_TYPE_PATH = "extr-data-phen-column-to-type.json"
    EXTR_CLINICAL_UNITS_PATH = "extr-data-clin-column-to-unit.json"
    EXTR_PHENOTYPIC_DOMAIN_PATH = "extr-data-phen-column-to-domain.json"
    EXTR_CLINICAL_DOMAIN_PATH = "extr-data-clin-column-to-domain.json"

    PIDS_PATH = "anonymized_patient_ids.json"
    # EXTR_FILLED_PIDS_PATH = "extr-filled-pids.json"



import json

from database.Database import Database
from database.Execution import Execution
from enums.TableNames import TableNames
from etl.Task import Task
from main_statistics.DatabaseStatistics import DatabaseStatistics
from main_statistics.QualityStatistics import QualityStatistics
from main_statistics.TimeStatistics import TimeStatistics
from utils.setup_logger import log


class Reporting(Task):

    NB_INSTANCES = "Number of instances"
    REC_NO_VALUE = "Record instances with no 'value' field"
    REC_NO_VALUE_INST = "Record instances with no 'value' field per 'instantiate'"
    CC_EMPTY_TEXT = "CodeableConcept with empty 'text' field"
    CC_EMPTY_LIST = "CodeableConcept with empty 'list' field"

    def __init__(self, database: Database, execution: Execution, quality_stats: QualityStatistics, db_stats: DatabaseStatistics,
                 time_stats: TimeStatistics):
        super().__init__(database=database, execution=execution, quality_stats=QualityStatistics(record_stats=False))
        self.quality_stats = quality_stats
        self.db_stats = db_stats
        self.time_stats = time_stats
        log.info(self.time_stats)

        # to get a user-friendly report
        self.report = {}

        # counts over the database after the ETL has finished
        self.counts_instances = {}
        self.nb_of_rec_with_no_value = {}
        self.nb_of_rec_with_no_value_per_instantiate = {}
        self.nb_of_cc_with_no_text_per_table = {}
        self.nb_of_cc_with_no_onto_resources_per_table = {}

    def run(self):
        self.report = {}
        self.report = self.report | self.db_stats.to_json()
        self.report = self.report | self.quality_stats.to_json()
        self.report = self.report | self.time_stats.to_json()

        # 4. print the final report
        # self.print_report()

        # 5. save each stat report in the database
        self.database.drop_table(table_name=TableNames.STATS_TIME)
        self.database.insert_one_tuple(table_name=TableNames.STATS_DB, one_tuple=self.db_stats.to_json())
        self.database.insert_one_tuple(table_name=TableNames.STATS_TIME, one_tuple=self.time_stats.to_json())
        self.database.insert_one_tuple(table_name=TableNames.STATS_QUALITY, one_tuple=self.quality_stats.to_json())

    def print_report(self) -> None:
        log.info("**** FINAL REPORT ****")
        # to get double-quoted keys and check it in jsonlint
        # we use default=str to be able to serialize datetime (otherwise non-serializable object)
        log.info(json.dumps(self.report, default=str))



import dataclasses

from main_statistics.MainStatistics import MainStatistics


@dataclasses.dataclass()
class QualityStatistics(MainStatistics):
    columns_no_ontology: list = dataclasses.field(default_factory=list)  # list of column names for which no ontology resource is provided
    columns_no_etl_type: list = dataclasses.field(default_factory=list)  # list of column names for which no etl type is provided
    columns_unmatched_typeof_etl_types: dict = dataclasses.field(default_factory=dict)  # { column_name: { "typeof_type": ttype, "etl_type": etype }, ... }
    columns_unknown_ontology: dict = dataclasses.field(default_factory=dict)  # { column_name: [set of unknown ontology names], ... }
    columns_unknown_etl_type: dict = dataclasses.field(default_factory=dict)  # { column_name: unknown_etl_type, ... }
    categorical_columns_without_json_values: list = dataclasses.field(default_factory=list)  # list of column names for which the ETL type is "category" and for which the column JSON_values is empty
    categorical_columns_unparseable_json: dict = dataclasses.field(default_factory=dict)  # { column_name: broken_json, ... }
    diagnosis_no_orphanet_code: list = dataclasses.field(default_factory=list)  # list of diagnosis standard names for which no orphanet code is provided
    data_columns_not_in_metadata: list = dataclasses.field(default_factory=list)  # list of data columns that are "removed" because not part of the metadata
    empty_cells_per_column: dict = dataclasses.field(default_factory=dict)  # { "column_name": X }
    failed_api_calls: dict = dataclasses.field(default_factory=dict)  # { "ontology/id_code": api_error, ... }
    unknown_categorical_values: dict = dataclasses.field(default_factory=dict)  # { column_name: [ set of unknown categorical values ], ... }
    unknown_boolean_values: dict = dataclasses.field(default_factory=dict)  # { column_name: [ set of unknown categorical boolean values ], ... }
    numerical_values_unmatched_unit: dict = dataclasses.field(default_factory=dict)  # { column_name: { value: { "expected_dim": exp_dim, "current_unit": curr_unit }, ... }, ... }
    non_numeric_values_with_unit: dict = dataclasses.field(default_factory=dict)  # { column_name: { value: curr_dim, ... }, ... }
    
    def add_column_with_no_ontology(self, column_name: str):
        if self.record_stats and column_name not in self.columns_no_ontology:
            self.columns_no_ontology.append(column_name)

    def add_column_with_no_etl_type(self, column_name: str):
        if self.record_stats and column_name not in self.columns_no_etl_type:
            self.columns_no_etl_type.append(column_name)

    def add_column_with_unmatched_typeof_etl_types(self, column_name: str, typeof_type: str, etl_type: str):
        # {column_name: { etl_type: [all encountered var types different from etl_type] }, ... }
        if self.record_stats:
            if column_name not in self.columns_unmatched_typeof_etl_types:
                self.columns_unmatched_typeof_etl_types[column_name] = {etl_type: []}
            if typeof_type not in self.columns_unmatched_typeof_etl_types[column_name][etl_type]:
                self.columns_unmatched_typeof_etl_types[column_name][etl_type].append(typeof_type)

    def add_column_unknown_ontology(self, column_name: str, ontology_name: str):
        if self.record_stats:
            if column_name not in self.columns_unknown_ontology:
                self.columns_unknown_ontology[column_name] = []
            if ontology_name not in self.columns_unknown_ontology[column_name]:
                self.columns_unknown_ontology[column_name].append(ontology_name)

    def add_column_unknown_etl_type(self, column_name: str, etl_type: str):
        if self.record_stats and column_name not in self.columns_unknown_etl_type:
            self.columns_unknown_etl_type[column_name] = etl_type

    def add_categorical_column_with_no_json(self, column_name: str):
        if self.record_stats and column_name not in self.categorical_columns_without_json_values:
            self.categorical_columns_without_json_values.append(column_name)

    def add_categorical_colum_with_unparseable_json(self, column_name: str, broken_json: str):
        if self.record_stats and column_name not in self.categorical_columns_unparseable_json:
            self.categorical_columns_unparseable_json[column_name] = broken_json

    def add_diagnosis_with_no_orphanet_code(self, diagnosis_standard_name: str):
        if self.record_stats and diagnosis_standard_name not in self.diagnosis_no_orphanet_code:
            self.diagnosis_no_orphanet_code.append(diagnosis_standard_name)

    def add_column_not_described_in_metadata(self, data_column_name: str):
        if self.record_stats and data_column_name not in self.data_columns_not_in_metadata:
            self.data_columns_not_in_metadata.append(data_column_name)

    def count_empty_cell_for_column(self, column_name: str) -> None:
        if column_name not in self.empty_cells_per_column:
            self.empty_cells_per_column[column_name] = 1
        else:
            self.empty_cells_per_column[column_name] += 1

    def add_failed_api_call(self, system: str, code: str, api_error: str):
        if self.record_stats and f"{system}/{code}" not in self.failed_api_calls:
            self.failed_api_calls[f"{system}/{code}"] = api_error

    def add_unknown_categorical_value(self, column_name: str, categorical_value: str):
        if self.record_stats:
            if column_name not in self.unknown_categorical_values:
                self.unknown_categorical_values[column_name] = []
            if categorical_value not in self.unknown_categorical_values[column_name]:
                self.unknown_categorical_values[column_name].append(categorical_value)

    def add_unknown_boolean_value(self, column_name: str, boolean_value: str):
        if self.record_stats:
            if column_name not in self.unknown_boolean_values:
                self.unknown_boolean_values[column_name] = []
            if boolean_value not in self.unknown_boolean_values[column_name]:
                self.unknown_boolean_values[column_name].append(boolean_value)

    def add_numerical_value_with_unmatched_unit(self, column_name: str, expected_unit: str, current_unit: str, value: str):
        # { column_name: { value: { "expected_dim": exp_dim, "current_unit": curr_unit }, ... }, ... }
        if self.record_stats:
            if column_name not in self.numerical_values_unmatched_unit:
                self.numerical_values_unmatched_unit[column_name] = {}
            if value not in self.numerical_values_unmatched_unit[column_name]:
                self.numerical_values_unmatched_unit[column_name][value] = {}
            self.numerical_values_unmatched_unit[column_name][value] = {"expected_unit": expected_unit, "current_unit": current_unit}

    def add_non_numeric_value_with_unit(self, column_name: str, unit: str, value: str):
        # { column_name: { value: curr_dim, ... }, ... }
        if self.record_stats:
            if column_name not in self.non_numeric_values_with_unit:
                self.non_numeric_values_with_unit[column_name] = {}
            if value not in self.non_numeric_values_with_unit[column_name]:
                self.non_numeric_values_with_unit[column_name][value] = unit



import os
from datetime import timedelta, datetime
from typing import Any

from database.Operators import THE_DATETIME_FORMAT
from entities.Record import Record
from entities.Resource import Resource
from utils.setup_logger import log


# for setting up the tests with specific (env) parameters
def set_env_variables_from_dict(env_vars: dict):
    # it is not possible to change the values of a .env and to write them in the .env file
    # so, we simply change them within the env
    for key, value in env_vars.items():
        os.environ[key] = value


# for comparing keys in original and inserted objects
def compare_keys(original_object: dict, inserted_object: dict) -> bool:
    original_keys = list(original_object.keys())
    inserted_keys = list(inserted_object.keys())
    if "_id" in inserted_keys:
        inserted_keys.remove("_id")
    # we need to sort the keys, otherwise the equality would not work, e.g., ["a", "b"] != ["b", "a"]
    inserted_keys.sort()
    original_keys.sort()

    return inserted_keys == original_keys


# for comparing two tuples
def compare_tuples(original_tuple: dict, inserted_tuple: dict) -> None:
    assert inserted_tuple is not None, "The inserted tuple is None."
    assert compare_keys(original_object=original_tuple, inserted_object=inserted_tuple) is True, different_keys()
    for original_key in original_tuple:
        assert original_key in inserted_tuple, missing_attribute(original_key)
        if original_key == Resource.TIMESTAMP_:
            # for the special case of timestamp attributes, we check +/- few milliseconds
            # to avoid failing when conversions do not yield the exact same datetime value
            inserted_time = datetime.strptime(inserted_tuple[original_key]["$date"], THE_DATETIME_FORMAT)
            inserted_time_minus = inserted_time - timedelta(milliseconds=100)
            inserted_time_plus = inserted_time + timedelta(milliseconds=100)
            assert inserted_time_minus <= inserted_time <= inserted_time_plus
        else:
            assert original_tuple[original_key] == inserted_tuple[original_key], different_values(original_key)


# for logging when assert fails in test
def missing_attribute(attribute: str) -> str:
    return f"The inserted tuple is missing the attribute '{attribute}'."


def different_values(attribute: str) -> str:
    return f"The value for '{attribute}' differs between the original and the inserted tuples."


def different_keys() -> str:
    return "The keys of the original and inserted tuple differ."


def wrong_number_of_docs(number: int):
    return f"The expected number of documents is {number}."


def get_feature_by_text(features: list, feature_text: str) -> dict:
    """
    """
    for json_feature in features:
        if "name" in json_feature and json_feature["name"] == feature_text:
            return json_feature
    return {}


def get_records_for_patient(records: list, patient_id: str) -> list[dict]:
    """
    :param records: list of LabRecord resources
    """
    matching_records = []
    for json_record in records:
        if json_record[Record.SUBJECT_] == patient_id:
            matching_records.append(json_record)
    # also sort them by PhenFeature reference id
    log.info(matching_records)
    return sorted(matching_records, key=lambda d: d[Record.INSTANTIATES_])


def get_field_value_for_patient(records: list, features: list, patient_id: str, column_name: str) -> Any:
    """
    :param records: list of XRecord resources
    :param features: list of XFeature resource
    :param patient_id: the patient (ID) for which we want to get a specific value
    :param column_name: the column for which we want to get the value
    """

    # log.info(f"looking for the value of column {column_name} for patient {patient_id}")

    feature = None
    for one_feature in features:
        if one_feature["name"] == column_name:
            feature = one_feature
            break
    if feature is not None:
        for record in records:
            # log.info(f"checking {json_lab_record['has_subject']} vs. {patient_id} and {json_lab_record['instantiates']} vs. {feature['identifier']}")
            if record[Record.SUBJECT_] == patient_id and record[Record.INSTANTIATES_] == feature[Resource.IDENTIFIER_]:
                log.info(f"for patient {patient_id} and column {column_name}, record is {record}")
                return record[Record.VALUE_]
    return None



import dataclasses
import json

from constants.methods import factory
from database.Database import Database
from utils.setup_logger import log


@dataclasses.dataclass()
class Counter:
    resource_id: int = dataclasses.field(init=False, default=0)

    def increment(self) -> int:
        self.resource_id = self.resource_id + 1
        return self.resource_id

    def set(self, new_value) -> None:
        self.resource_id = new_value

    def set_with_database(self, database: Database) -> None:
        max_value = database.get_max_resource_counter_id() + 1  # start 1 after the current counter to avoid resources with the same ID
        if max_value > -1:
            self.set(max_value)
            log.debug(f"The resource counter is now set to {self.resource_id}.")

    def reset(self) -> None:
        self.resource_id = 0

    def to_json(self):
        return dataclasses.asdict(self, dict_factory=factory)

    def __str__(self):
        return json.dumps(self.to_json())



import dataclasses

from entities.Record import Record
from enums.Profile import Profile
from enums.TableNames import TableNames


@dataclasses.dataclass()
class ImagingRecord(Record):
    scan: str
    entity_type: str = f"{Profile.IMAGING}{TableNames.RECORD}"



from database.Database import Database
from database.Execution import Execution
from main_statistics.QualityStatistics import QualityStatistics
from main_statistics.TimeStatistics import TimeStatistics


class Task:
    def __init__(self, database: Database, execution: Execution,
                 quality_stats: QualityStatistics):
        self.database = database
        self.execution = execution
        self.quality_stats = quality_stats

    def run(self):
        raise NotImplementedError("Each class which inherits from Task should implement a run() method.")



from enums.EnumAsClass import EnumAsClass


class ParameterKeys(EnumAsClass):
    # this corresponds to (almost) all the keys defined in the .env
    HOSPITAL_NAME = "HOSPITAL_NAME"
    DB_NAME = "DB_NAME"
    DB_DROP = "DB_DROP"
    USE_LOCALE = "USE_LOCALE"
    COLUMNS_TO_REMOVE_KEY = "COLUMNS_TO_REMOVE"
    METADATA_PATH = "METADATA"
    DATA_FILES = "DATA_FILES"
    ANONYMIZED_PATIENT_IDS = "ANONYMIZED_PIDS"
    RECORD_CARRIER_PATIENT = "RECORD_CARRIER_PATIENTS"
    PATIENT_ID_COLUMN = "PATIENT_ID"
    SAMPLE_ID_COLUMN = "SAMPLE_ID"



import dataclasses

from entities.Resource import Resource
from enums.TableNames import TableNames


@dataclasses.dataclass()
class ResourceTest(Resource):
    entity_type: str = f"{TableNames.TEST}"



import dataclasses
import os.path
import uuid
from datetime import datetime

from catalogue.DatasetProfile import DatasetProfile
from constants.defaults import DATASET_GLOBAL_IDENTIFIER_PREFIX
from database.Database import Database
from entities.Resource import Resource
from enums.TableNames import TableNames
from utils.setup_logger import log


@dataclasses.dataclass()
class Dataset(Resource):
    database: Database = dataclasses.field(repr=False)
    docker_path: str
    global_identifier: str = dataclasses.field(init=False)
    version: str = dataclasses.field(init=False)
    release_date: datetime.date = dataclasses.field(init=False)
    last_update: datetime.date = dataclasses.field(init=False)
    version_notes: str
    license: str
    profile: DatasetProfile = dataclasses.field(init=False)

    # keys to be used when writing JSON or queries
    # those names have to exactly match the variables names declared in entity classes
    GID_ = "global_identifier"

    # keys to be used when using an entity attribute as a query variable
    GID__ = f"${GID_}"

    def __post_init__(self):
        super().__post_init__()
        log.info(self.docker_path)
        from_database = False
        if self.database is not None:
            results = self.database.find_operation(table_name=TableNames.DATASET, filter_dict={"docker_path": self.docker_path}, projection={})
            for result in results:
                log.info(result)
                # there was a dataset
                self.global_identifier = result["global_identifier"]
                log.info(f"existing dataset identifier: {self.global_identifier}")
                self.version = str(int(result["version"]) + 1)  # increment the existing dataset version
                self.last_update = datetime.now()  # the release should be only computed the first time the dataset is inserted
                self.version_notes = result["version_notes"] if "version_notes" in result else None
                self.license = result["license"] if "license" in result else None
                from_database = True
        if not from_database:
            # no dataset was corresponding, initializing all fields
            self.global_identifier = Dataset.compute_global_identifier()
            log.info(f"new dataset identifier: {self.global_identifier}")
            self.version = "1"  # computed by incrementing the previous version (obtained from the db) - starts at 1
            self.release_date = datetime.now()
            self.last_update = datetime.now()
        log.info(self.global_identifier)
        # we get the size in Mb and round it to 6 digits
        size = round(os.path.getsize(self.docker_path)/(1000*1000), 6)
        self.profile = DatasetProfile(description="", theme="",
                                      filetype=self.get_file_type(self.docker_path), size=size, nb_tuples=0,
                                      completeness=0, uniqueness=0)

    @classmethod
    def get_file_type(cls, docker_path) -> str:
        _, file_extension = os.path.splitext(docker_path)
        return file_extension[1:]  # remove the comma kept at the beginning of the extension

    @classmethod
    def compute_global_identifier(cls) -> str:
        return DATASET_GLOBAL_IDENTIFIER_PREFIX + str(uuid.uuid4())



import bson
import pandas as pd
import pymongo
from pymongo.mongo_client import MongoClient

from enums.DataTypes import DataTypes
from enums.MetadataColumns import MetadataColumns


def main_load_json_from_file_as_bson():
    with MongoClient() as mongo:
        db = mongo.get_database("mydb")
        with open("../datasets/test_dates/data.json", "r") as data_file:
            my_tuples = bson.json_util.loads(data_file.read())
            db["mycollectionIX"].insert_many(my_tuples, ordered=False)


def main_build_upsert():
    tuples = [{"a": 1, "b": 2, "c": 3}, {"a": 4, "b": 5, "d": 6}]
    unique_variables = ["a", "d"]

    operations = [pymongo.UpdateOne(
        filter={unique_variable: one_tuple[unique_variable] for unique_variable in unique_variables if unique_variable in one_tuple},
        update={"$set": one_tuple}, upsert=True) for one_tuple in tuples
    ]
    print(operations)


def main_na_pandas():
    na_values = ["no information", "No information", "No Information", "-", "0", "0.0", "-0", "-0.0"]
    df = pd.read_csv("/Users/nelly/Documents/boulot/postdoc-polimi/etl/datasets/test/orig-data-clin.csv", index_col=False, dtype=str, na_values=[], keep_default_na=False)
    print(df)
    print(df.columns)

    print(df)
    for column in df:
        df.loc[:, column] = df[column].apply(lambda x: MetadataColumns.normalize_value(column_value=x))
    print(df)

    for row in df.itertuples(index=False):
        print(row)
        for column_name in df.columns:
            value = row[df.columns.get_loc(column_name)]
            print(f"      '{value}' (type={type(value)}, None={value is None}, Null={pd.isnull(value)}, empty={value == ""})")


def main_build_mongodb_types():
    # key is the Mongodb type, value is the ETL type
    reverse_mapping = {
        "int": DataTypes.INTEGER,
        "double": DataTypes.FLOAT,
        "string": DataTypes.STRING,
        "date": DataTypes.DATE,
        "timestamp": DataTypes.DATETIME,
        "bool": DataTypes.BOOLEAN,
        "object": DataTypes.CATEGORY,
        # "string": Datatypes.IMAGE,
        # "string": Datatypes.REGEX,
        # "string": Datatypes.API,
        # "string": Datatypes.LIST,
    }
    type_switch = {
        "$switch": {
            "branches": [
                {
                    "case": {"$eq": ["$_id.featureType", bson_type]},
                    "then": user_alias
                }
                for bson_type, user_alias in reverse_mapping.items()
            ],
            "default": "unknown"
        }
    }
    print(type_switch)


if __name__ == '__main__':
    main_build_mongodb_types()
    print("Done.")



import json
import os
import unittest

import numpy as np
import pandas as pd

from constants.structure import TEST_DB_NAME, DOCKER_FOLDER_TEST
from database.Database import Database
from database.Execution import Execution
from enums.DataTypes import DataTypes
from enums.HospitalNames import HospitalNames
from enums.MetadataColumns import MetadataColumns
from enums.Ontologies import Ontologies
from enums.ParameterKeys import ParameterKeys
from enums.Profile import Profile
from enums.TheTestFiles import TheTestFiles
from etl.Extract import Extract
from main_statistics.QualityStatistics import QualityStatistics
from utils.file_utils import read_tabular_file_as_string
from utils.test_utils import set_env_variables_from_dict


# personalized setup called at the beginning of each test
def my_setup(metadata_path: str, data_paths: str, data_type: str, pids_path: str, hospital_name: str) -> Extract:
    args = {
        ParameterKeys.HOSPITAL_NAME: hospital_name,
        ParameterKeys.DB_NAME: TEST_DB_NAME,
        ParameterKeys.DB_DROP: "True",
        ParameterKeys.METADATA_PATH: metadata_path,
        ParameterKeys.DATA_FILES: data_paths,
        ParameterKeys.ANONYMIZED_PATIENT_IDS: pids_path
    }
    set_env_variables_from_dict(env_vars=args)
    TestExtract.execution.internals_set_up()
    TestExtract.execution.file_set_up(setup_files=True)  # we need to set up the metadata file at least
    # in the dev mode, the current_filepath variable is set during the ETL
    # for tests, we need to manually set it because we have no ETL instance
    TestExtract.execution.current_filepath = os.path.join(DOCKER_FOLDER_TEST, data_paths)
    metadata = read_tabular_file_as_string(filepath=os.path.join(DOCKER_FOLDER_TEST, metadata_path))
    database = Database(execution=TestExtract.execution)
    extract = Extract(metadata=metadata, profile=str(data_type), database=database, execution=TestExtract.execution, quality_stats=QualityStatistics(record_stats=False))
    extract.filter_metadata_file()
    extract.normalize_metadata_file()
    extract.load_tabular_data()
    extract.pre_process_data_file()
    extract.filter_data_file()
    extract.normalize_data_file()
    return extract


class TestExtract(unittest.TestCase):
    execution = Execution()

    def test_load_metadata_file_H1_D1(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_CLINICAL_PATH,
                           data_type=Profile.CLINICAL,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)

        # a. general size checks
        assert extract.metadata is not None, "Metadata is None, while it should not."
        assert len(extract.metadata.columns) == 21, "The expected number of columns is 21."
        assert len(extract.metadata) == 6, "The expected number of lines is 6."

        # b. checking the "sid" line completely
        assert extract.metadata[MetadataColumns.ONTO_NAME][1] == ""
        assert extract.metadata[MetadataColumns.ONTO_CODE][1] == ""
        assert extract.metadata[MetadataColumns.COLUMN_NAME][1] == "sid"  # all lower case
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][1] == "The Sample ID"  # kept as it is in the metadata for more clarity
        assert extract.metadata[MetadataColumns.ETL_TYPE][1] == DataTypes.STRING
        assert extract.metadata[MetadataColumns.JSON_VALUES][1] == ""
        assert extract.metadata[MetadataColumns.VISIBILITY][1] == "PUBLIC"
        assert extract.metadata[MetadataColumns.DOMAIN][1] == ""

        # g. more general checks
        # DATASET: this should be the dataset name, and there should be no other datasets in that column
        unique_dataset_names = list(extract.metadata[MetadataColumns.DATASET_NAME].unique())
        assert len(unique_dataset_names) == 1
        assert unique_dataset_names[0] == TheTestFiles.ORIG_CLINICAL_PATH.split(os.sep)[-1]

    def test_load_metadata_file_H1_D2(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_PHENOTYPIC_PATH,
                           data_type=Profile.PHENOTYPIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)

        # a. general size checks
        assert extract.metadata is not None, "Metadata is None, while it should not."
        assert len(extract.metadata.columns) == 21, "The expected number of columns is 21."
        assert len(extract.metadata) == 4, "The expected number of lines is 4."

        # b. checking the "sex" line completely
        assert extract.metadata[MetadataColumns.ONTO_NAME][1] == "snomedct"
        assert extract.metadata[MetadataColumns.ONTO_CODE][1] == "123: 789"
        assert extract.metadata[MetadataColumns.COLUMN_NAME][1] == "sex"  # all lower case
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][1] == "The sex at birth"  # kept as it is in the metadata for more clarity
        # pandas dataframe does not allow json objects, so we have to store them as JSON-like string
        expected_json_values = [{"value": "M", "snomedct": "248153007"}, {"value": "F", "snomedct": "248152002"}]
        # "[{""value"": ""M"", ""snomedct"": ""248153007""}, {""value"": ""F"", ""snomedct"": ""248152002""}]"
        assert extract.metadata[MetadataColumns.JSON_VALUES][1] == json.dumps(expected_json_values)
        assert extract.metadata[MetadataColumns.VISIBILITY][1] == "PUBLIC"
        assert extract.metadata[MetadataColumns.DOMAIN][1] == ""

        # c. test the first (Patient ID) line, because there are no ontologies for this one
        assert extract.metadata[MetadataColumns.COLUMN_NAME][0] == "id"  # normalized column name
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][0] == "The Patient ID"  # non-normalized description

        # d. test normalization of ontology names
        assert extract.metadata[MetadataColumns.ONTO_NAME][0] == ""
        assert extract.metadata[MetadataColumns.ONTO_NAME][1] == "snomedct"
        assert extract.metadata[MetadataColumns.ONTO_NAME][2] == "loinc"
        assert extract.metadata[MetadataColumns.ONTO_NAME][3] == ""

        # assert that codes are not normalized yet (they will be normalized within OntoResource)
        assert extract.metadata[MetadataColumns.ONTO_CODE][0] == ""
        assert extract.metadata[MetadataColumns.ONTO_CODE][1] == "123: 789"
        assert extract.metadata[MetadataColumns.ONTO_CODE][2] == " 46463-6 "
        assert extract.metadata[MetadataColumns.ONTO_CODE][3] == ""

        # e. check ETL type normalization
        assert extract.metadata[MetadataColumns.ETL_TYPE][0] == DataTypes.INTEGER  # patient id
        assert extract.metadata[MetadataColumns.ETL_TYPE][1] == DataTypes.CATEGORY  # sex
        assert extract.metadata[MetadataColumns.ETL_TYPE][2] == DataTypes.STRING  # ethnicity
        assert extract.metadata[MetadataColumns.ETL_TYPE][3] == DataTypes.DATETIME  # date of birth

        # f. more general checks
        # DATASET: this should be the dataset name, and there should be no other datasets in that column
        unique_dataset_names = list(extract.metadata[MetadataColumns.DATASET_NAME].unique())
        assert len(unique_dataset_names) == 1
        assert unique_dataset_names[0] == TheTestFiles.ORIG_PHENOTYPIC_PATH.split(os.sep)[-1]

    def test_load_metadata_file_H1_D3(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_DIAGNOSIS_PATH,
                           data_type=Profile.DIAGNOSIS,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)

        # a. general size checks
        assert extract.metadata is not None, "Metadata is None, while it should not."
        assert len(extract.metadata.columns) == 21, "The expected number of columns is 21."
        assert len(extract.metadata) == 3, "The expected number of lines is 3."

        # b. checking the first line completely
        assert extract.metadata[MetadataColumns.ONTO_NAME][2] == "omim"
        assert extract.metadata[MetadataColumns.ONTO_CODE][2] == "1245/   983 "  # this will be normalized later when building OntologyResource objects
        assert extract.metadata[MetadataColumns.COLUMN_NAME][2] == "disease_form"  # all lower case, with an underscore
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][2] == "The form of the disease"  # kept as it is in the metadata for more clarity
        assert extract.metadata[MetadataColumns.ETL_TYPE][2] == "category"  # all lower case
        # test JSON values in the second line (disease form)
        # pandas dataframe does not allow json objects, so we have to store them as JSON-like string
        # expected_json_values = [{"value": "start", "pubchem": "023468"}, {"value": "middle", "pubchem": "203:468"}, {"value": "end", "pubchem": "4097625"}]
        # assert extract.metadata[MetadataColumns.JSON_VALUES][2] == json.dumps(expected_json_values)

        # c. test the first (Patient ID) line, because there are no ontologies for this one
        assert extract.metadata[MetadataColumns.COLUMN_NAME][0] == "id"  # normalized column name
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][0] == "The Patient ID"  # non-normalized description

        # d. test normalization of ontology names
        assert extract.metadata[MetadataColumns.ONTO_NAME][1] == "omim"
        assert extract.metadata[MetadataColumns.ONTO_NAME][2] == "omim"
        assert extract.metadata[MetadataColumns.ONTO_NAME][0] == ""

        # d. test normalization of ontology codes (later normalized)
        assert extract.metadata[MetadataColumns.ONTO_CODE][1] == "1569 - 456"  # this will be normalized later with OntologyResource
        assert extract.metadata[MetadataColumns.ONTO_CODE][2] == "1245/   983 "
        assert extract.metadata[MetadataColumns.ONTO_CODE][0] == ""

        # e. more general checks
        # DATASET: this should be the dataset name, and there should be no other datasets in that column
        unique_dataset_names = list(extract.metadata[MetadataColumns.DATASET_NAME].unique())
        assert len(unique_dataset_names) == 1
        assert unique_dataset_names[0] == TheTestFiles.ORIG_DIAGNOSIS_PATH.split(os.sep)[-1]

    def test_load_metadata_file_H3_D1(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_GENOMICS_PATH,
                           data_type=Profile.GENOMIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H3)

        # a. general size checks
        assert extract.metadata is not None, "Metadata is None, while it should not."
        assert len(extract.metadata.columns) == 21, "The expected number of columns is 21."
        assert len(extract.metadata) == 3, "The expected number of lines is 3."

        # b. checking the first line completely
        assert extract.metadata[MetadataColumns.ONTO_NAME][2] == "loinc"
        assert extract.metadata[MetadataColumns.ONTO_CODE][2] == "3265970"
        assert extract.metadata[MetadataColumns.COLUMN_NAME][2] == "is_inherited"  # all lower case, with an underscore
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][2] == "Whether the gene is inherited"  # kept as it is in the metadata for more clarity
        assert extract.metadata[MetadataColumns.ETL_TYPE][2] == "bool"  # all lower case
        assert extract.metadata[MetadataColumns.JSON_VALUES][2] == ""

        # c. test the first (Patient ID) line, because there are no ontologies for this one
        assert extract.metadata[MetadataColumns.COLUMN_NAME][0] == "id"  # normalized column name
        assert extract.metadata[MetadataColumns.SIGNIFICATION_EN][0] == "The Patient ID"  # non-normalized description

        # d. test normalization of ontology codes
        assert extract.metadata[MetadataColumns.ONTO_NAME][1] == "loinc"
        assert extract.metadata[MetadataColumns.ONTO_NAME][2] == "loinc"
        assert extract.metadata[MetadataColumns.ONTO_NAME][0] == ""

        assert extract.metadata[MetadataColumns.ONTO_CODE][1] == "326597056"
        assert extract.metadata[MetadataColumns.ONTO_CODE][2] == "3265970"
        assert extract.metadata[MetadataColumns.ONTO_CODE][0] == ""

        # e. more general checks
        # DATASET: this should be the dataset name, and there should be no other datasets in that column
        unique_dataset_names = list(extract.metadata[MetadataColumns.DATASET_NAME].unique())
        assert len(unique_dataset_names) == 1
        assert unique_dataset_names[0] == TheTestFiles.ORIG_GENOMICS_PATH.split(os.sep)[-1]

    def test_load_data_file_H1_D1(self):
        extract = my_setup(metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                           data_paths=TheTestFiles.ORIG_CLINICAL_PATH,
                           data_type=Profile.CLINICAL,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        # a. general size checks
        assert extract.data is not None, "Data is None, while it should not."
        assert len(extract.data.columns) == 5, "The expected number of columns is 5."  # because molecule_y is in the metadata but not the data
        assert len(extract.data) == 10, "The expected number of lines is 10."

        # b. checking the first line completely
        # recall that in dataframes (and Pandas):
        # - everything is a string (we will cast them to their true type when building resources in the Transform step)
        # - when a column contains at least one empty cell, the column type is float, thus numbers are read as floats
        #   for instance 100 is read 100 if there are no empty cells in that column, 100.0 otherwise
        assert extract.data["sid"][0] == "s1", "The expected id is 's1'."
        assert extract.data["id"][0] == "999999999", "The expected id is '999999999'."
        assert extract.data["molecule_a"][0] == "0.001", "The expected value is '0.001'."
        assert extract.data["molecule_b"][0] == "100g", "The expected value is '100'."
        assert extract.data["molecule_g"][0] == "1", "The expected value is '1'."  # this will be later converted to bool
        # assert extract.data["molecule_z"][0] == "abc", "The expected value is 'abc'."  # this has been removed because not described in the metadata

    def test_load_data_file_H1_D2(self):
        extract = my_setup(metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                           data_paths=TheTestFiles.ORIG_PHENOTYPIC_PATH,
                           data_type=Profile.PHENOTYPIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)

        # a. general size checks
        assert extract.data is not None, "Data is None, while it should not."
        assert len(extract.data.columns) == 4, "The expected number of columns is 4."
        assert len(extract.data) == 10, "The expected number of lines is 10."

        # b. checking the first line completely
        # recall that in dataframes (and Pandas):
        # - everything is a string (we will cast them to their true type when building resources in the Transform step)
        # - when a column contains at least one empty cell, the column type is float, thus numbers are read as floats
        #   for instance 100 is read 100 if there are no empty cells in that column, 100.0 otherwise
        assert extract.data["id"][0] == "999999999", "The expected id is '999999999'."
        assert extract.data["sex"][0] == "f"
        assert extract.data["ethnicity"][0] == "white"
        assert extract.data["date_of_birth"][0] == ""

    def test_load_data_file_H3_D1(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_GENOMICS_PATH,
                           data_type=Profile.GENOMIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H3)

        # a. general size checks
        assert extract.data is not None, "Data is None, while it should not."
        assert len(extract.data.columns) == 3, "The expected number of columns is 3."
        assert len(extract.data) == 10, "The expected number of lines is 10."

        # b. checking the first line completely
        assert extract.data["id"][0] == "999999999"
        assert extract.data["gene"][0] == "abc-123"
        assert extract.data["is_inherited"][0] == "true"
        assert extract.data["gene"][1] == "abc-123"
        assert extract.data["is_inherited"][1] == "true"
        assert extract.data["gene"][2] == "abc-128"
        assert extract.data["is_inherited"][2] == "false"
        assert extract.data["gene"][3] == "abd-123"
        assert extract.data["is_inherited"][3] == "true"
        assert extract.data["gene"][4] == "ade-183"
        assert pd.isnull(extract.data["is_inherited"][4])  # the "n/a" value is indeed converted to a NaN
        assert extract.data["gene"][5] == "sdr-125"
        assert extract.data["is_inherited"][5] == "false"
        assert extract.data["gene"][6] == "dec-123"
        assert extract.data["is_inherited"][6] == "false"
        assert extract.data["gene"][7] == "gft-568"
        assert pd.isnull(extract.data["is_inherited"][7])  # the "NaN" value is indeed converted to a NaN
        assert extract.data["gene"][8] == "plo-719"
        assert extract.data["is_inherited"][8] == ""
        assert extract.data["gene"][9] == ""
        assert extract.data["is_inherited"][9] == ""

    def test_compute_sam_mapped_values(self):
        extract = my_setup(metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                           data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                           data_type=Profile.CLINICAL,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        extract.compute_mapping_categorical_value_to_onto_resource()

        print(extract.mapping_column_to_categorical_value)
        assert "molecule_g" in extract.mapping_column_to_categorical_value  # only one categorical feature
        assert len(extract.mapping_column_to_categorical_value) == 1
        assert len(extract.mapping_column_to_categorical_value["molecule_g"].keys()) == 3  # 0, 1, NA
        # checking "0" mapping
        assert "0" in extract.mapping_column_to_categorical_value["molecule_g"]  # normalized categorical value
        cc_0 = extract.mapping_column_to_categorical_value["molecule_g"]["0"]
        assert len(cc_0) == 3  # system, code, and label keys
        assert "system" in cc_0
        assert "code" in cc_0
        assert "label" in cc_0
        assert cc_0["label"] == "No"  # display got from the API
        assert cc_0["system"] == Ontologies.SNOMEDCT["url"]  # normalized (ontology) key
        assert cc_0["code"] == "373067005"  # normalized ontology code
        # checking "1" mapping
        assert "1" in extract.mapping_column_to_categorical_value["molecule_g"]  # normalized categorical value
        cc_1 = extract.mapping_column_to_categorical_value["molecule_g"]["1"]
        assert len(cc_1) == 3  # system, code, and label keys
        assert "system" in cc_1
        assert "code" in cc_1
        assert "label" in cc_1
        assert cc_1["label"] == "Yes"  # display computed with the API
        assert cc_1["system"] == Ontologies.SNOMEDCT["url"]  # normalized (ontology) key
        assert cc_1["code"] == "373066001"  # normalized ontology code
        # checking "na" mapping
        assert np.nan in extract.mapping_column_to_categorical_value["molecule_g"]  # normalized categorical value; np.nan is the key name
        cc_na = extract.mapping_column_to_categorical_value["molecule_g"][np.nan]
        assert len(cc_na) == 3  # system, code, and label keys
        assert "system" in cc_na
        assert "code" in cc_na
        assert "label" in cc_na
        assert cc_na["label"] == "Null"  # display computed with the API
        assert cc_na["system"] == Ontologies.SNOMEDCT["url"]  # normalized (ontology) key
        assert cc_na["code"] == "276727009"  # normalized ontology code

    def test_compute_phen_mapped_values(self):
        extract = my_setup(metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                           data_paths=TheTestFiles.EXTR_PHENOTYPIC_DATA_PATH,
                           data_type=Profile.PHENOTYPIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        extract.compute_mapping_categorical_value_to_onto_resource()

        print(extract.mapping_column_to_categorical_value)
        assert "sex" in extract.mapping_column_to_categorical_value
        assert len(extract.mapping_column_to_categorical_value.keys()) == 1  # only sex feature is here
        # checking "male" mapping
        assert "m" in extract.mapping_column_to_categorical_value["sex"]  # normalized categorical value
        cc_male = extract.mapping_column_to_categorical_value["sex"]["m"]
        assert len(cc_male) == 3  # system, code, and label keys
        assert "system" in cc_male
        assert "code" in cc_male
        assert "label" in cc_male
        assert cc_male["label"] == "Male"  # display got from the API
        assert cc_male["system"] == Ontologies.SNOMEDCT["url"]  # normalized (ontology) key
        assert cc_male["code"] == "248153007"  # normalized ontology code
        # checking "female" mapping
        assert "f" in extract.mapping_column_to_categorical_value["sex"]  # normalized categorical value
        cc_female = extract.mapping_column_to_categorical_value["sex"]["f"]
        assert len(cc_female) == 3  # system, code, and label keys
        assert "system" in cc_female
        assert "code" in cc_female
        assert "label" in cc_female
        assert cc_female["label"] == "Female"  # display computed with the API
        assert cc_female["system"] == Ontologies.SNOMEDCT["url"]  # normalized (ontology) key
        assert cc_female["code"] == "248152002"  # normalized ontology code

    def test_removed_unused_columns(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_CLINICAL_PATH,
                           data_type=Profile.CLINICAL,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)

        # we assert that we get rid of the 'molecule_z' column (because it was not described in the metadata)
        # other columns are kept
        remaining_data_columns = list(extract.data.columns)
        remaining_data_columns.sort()
        expected_columns = ["id", "molecule_a", "molecule_b", "molecule_g", "sid"]  # do not change order, otherwise the comparison fails
        assert len(remaining_data_columns) == len(expected_columns)  # molecule_z has been removed
        assert remaining_data_columns == expected_columns

        # we assert that we kept 'molecule_y' column (even though it was not in the data, but still described in the metadata)
        # other columns are kept
        described_columns = list(extract.metadata[MetadataColumns.COLUMN_NAME])
        described_columns.sort()
        expected_columns = ["id", "molecule_a", "molecule_b", "molecule_g", "molecule_y", "sid"]
        assert len(described_columns) == len(expected_columns)  # molecule_y has been kept
        assert described_columns == expected_columns

    def test_compute_clin_column_to_unit(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_CLINICAL_PATH,
                           data_type=Profile.CLINICAL,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        extract.compute_column_to_unit()

        # {'id': None, 'molecule_a': "mg/L", 'molecule_b': "g", 'molecule_g': None, 'molecule_z': None, "molecule_y": None}
        assert len(extract.mapping_column_to_unit.keys()) == 6
        assert "id" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["id"] is None
        assert "molecule_a" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["molecule_a"] == "mg/L"
        assert "molecule_b" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["molecule_b"] is None  # because we do not extract units from data anymore
        assert "molecule_g" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["molecule_g"] is None
        assert "molecule_y" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["molecule_y"] is None

    def test_compute_phen_column_to_unit(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_PHENOTYPIC_PATH,
                           data_type=Profile.PHENOTYPIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        extract.compute_column_to_unit()

        # {'id': None, 'sex': [], 'ethnicity': [], 'date_of_birth': []}
        assert len(extract.mapping_column_to_unit.keys()) == 4
        assert "id" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["id"] is None
        assert "sex" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["sex"] is None
        assert "ethnicity" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["ethnicity"] is None
        assert "date_of_birth" in extract.mapping_column_to_unit
        assert extract.mapping_column_to_unit["date_of_birth"] is None

    def test_compute_phen_column_to_domain(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_PHENOTYPIC_PATH,
                           data_type=Profile.PHENOTYPIC,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        extract.compute_column_to_domain()

        # {'id': None, 'sex': [], 'ethnicity': [], 'date_of_birth': []}
        assert len(extract.mapping_column_to_domain.keys()) == 4
        assert "id" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["id"] is None
        assert "sex" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["sex"] is None
        assert "ethnicity" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["ethnicity"] is None
        assert "date_of_birth" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["date_of_birth"] is None

    def test_compute_clin_column_to_domain(self):
        extract = my_setup(metadata_path=TheTestFiles.ORIG_METADATA_PATH,
                           data_paths=TheTestFiles.ORIG_CLINICAL_PATH,
                           data_type=Profile.CLINICAL,
                           pids_path=TheTestFiles.PIDS_PATH,
                           hospital_name=HospitalNames.TEST_H1)
        extract.compute_column_to_domain()

        # {'id': None,
        #  'sid': None,
        #  'molecule_a': {""min"": 0},
        #  'molecule_b': None,
        #  'molecule_g': None,
        #  'molecule_y': {""min"": 0, ""max"": 5}
        # }
        assert len(extract.mapping_column_to_domain.keys()) == 6
        assert "id" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["id"] is None
        assert "sid" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["sid"] is None
        assert "molecule_a" in extract.mapping_column_to_domain
        assert len(extract.mapping_column_to_domain["molecule_a"]) == 1
        assert "min" in extract.mapping_column_to_domain["molecule_a"]
        assert extract.mapping_column_to_domain["molecule_a"]["min"] == 0
        assert "molecule_b" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["molecule_b"] is None
        assert "molecule_g" in extract.mapping_column_to_domain
        assert extract.mapping_column_to_domain["molecule_g"] is None
        assert "molecule_y" in extract.mapping_column_to_domain
        assert len(extract.mapping_column_to_domain["molecule_y"]) == 2
        assert "min" in extract.mapping_column_to_domain["molecule_y"]
        assert extract.mapping_column_to_domain["molecule_y"]["min"] == 0
        assert "max" in extract.mapping_column_to_domain["molecule_y"]
        assert extract.mapping_column_to_domain["molecule_y"]["max"] == 5



from __future__ import annotations

import dataclasses
import getpass
import json
import logging
import os.path
import platform
from dataclasses import field
from datetime import datetime

import pymongo

from constants.methods import factory
from enums.MetadataColumns import MetadataColumns
from enums.Profile import Profile

from constants.structure import WORKING_DIR, DB_CONNECTION, DOCKER_FOLDER_METADATA, DOCKER_FOLDER_TEST, DEFAULT_DB_NAME
from enums.HospitalNames import HospitalNames
from enums.ParameterKeys import ParameterKeys
from utils import setup_logger
from utils.cast_utils import cast_str_to_int
from utils.setup_logger import log


@dataclasses.dataclass()
class Execution:
    execution_date: datetime = datetime.now()
    working_dir: str = field(init=False, default=os.path.join(os.getcwd(), WORKING_DIR))  # default in the code
    working_dir_current: str = field(init=False, default=None)  # computed in create_current_working_dir()

    # parameters related to the project structure and the input/output files
    metadata_filepath: str = field(init=False, default=None)  # user input
    current_filepath: str = field(init=False, default=None)  # set in the loop on files in ETL
    current_dataset_gid: str = field(init=False, default=None)  # set in the loop on file sin ETL
    current_file_number: int = field(init=False, default=1)  # set in the ETL
    anonymized_patient_ids_filepath: str = field(init=False, default=None)  # user input
    use_locale: str = field(init=False, default="en_GB")  # user input
    record_carrier_patients: bool = field(init=False, default=False)  # user input

    # parameters related to the database and the ETL
    hospital_name: str = field(init=False, default="")  # this will be given as input by users
    db_connection: str = field(init=False, default="mongodb://localhost:27018/")  # user input
    db_name: str = field(init=False, default=DEFAULT_DB_NAME)
    db_drop: bool = field(init=False, default=False)  # user input
    columns_to_remove: list = field(init=False, default_factory=list)  # user input
    patient_id_column_name: str = field(init=False, default="id")
    sample_id_column_name: str = field(init=False, default="")

    # parameters related to the execution context (python, pymongo, etc.)
    python_version: str = platform.python_version()
    pymongo_version: str = pymongo.version
    platform: str = platform.platform()
    user: str = getpass.getuser()

    def internals_set_up(self) -> None:
        log.info("in set_up")

        # set class variables using env. variables
        self.db_connection = DB_CONNECTION  # this is not a user parameter anymore because this is part of the Docker functioning, not something user should be able to change
        self.db_name = self.check_parameter(key=ParameterKeys.DB_NAME, accepted_values=None, default_value=self.db_name)
        log.debug(f"creating new DB with name {self.db_name}")
        log.info(HospitalNames.values())
        self.hospital_name = self.check_parameter(key=ParameterKeys.HOSPITAL_NAME, accepted_values=HospitalNames.values(), default_value=self.hospital_name)
        self.use_locale = self.check_parameter(key=ParameterKeys.USE_LOCALE, accepted_values=None, default_value=self.use_locale)
        self.db_drop = self.check_parameter(key=ParameterKeys.DB_DROP, accepted_values=["True", "False", True, False], default_value=self.db_drop)
        self.columns_to_remove = self.check_parameter(key=ParameterKeys.COLUMNS_TO_REMOVE_KEY, accepted_values=None, default_value=self.columns_to_remove)
        self.record_carrier_patients = self.check_parameter(key=ParameterKeys.RECORD_CARRIER_PATIENT, accepted_values=["True", "False", True, False], default_value=self.record_carrier_patients)
        self.patient_id_column_name = MetadataColumns.normalize_name(self.check_parameter(key=ParameterKeys.PATIENT_ID_COLUMN, accepted_values=None, default_value=self.patient_id_column_name))
        self.sample_id_column_name = MetadataColumns.normalize_name(self.check_parameter(key=ParameterKeys.SAMPLE_ID_COLUMN, accepted_values=None, default_value=self.sample_id_column_name))

        # create working files for the ETL
        self.create_current_working_dir()
        self.setup_logging_files()

    def file_set_up(self, setup_files: bool) -> None:
        log.info("in file_set_up()")

        # # D. set up the anonymized patient id data file
        # # this should NOT be merged with the setup of data files as this as to be set even though no data is provided
        # # (this happens in tests: data is given by hand, i.e., without set_up, but the anonymized patient IDs file still has to exist
        # self.anonymized_patient_ids_filepath = self.check_parameter(key=ParameterKeys.ANONYMIZED_PATIENT_IDS, accepted_values=None, default_value=self.anonymized_patient_ids_filepath)
        # log.debug(self.anonymized_patient_ids_filepath)
        # if self.anonymized_patient_ids_filepath is not None:
        #     # it may be None when we are computing the catalogue data (because we don't need it)
        self.setup_mapping_to_anonymized_patient_ids()

        # E. set up the data and metadata files
        if setup_files:
            log.debug("I will also set up data files")
            # 1. compute the (Docker-rooted) absolute path to the metadata file
            self.metadata_filepath = self.check_parameter(key=ParameterKeys.METADATA_PATH, accepted_values=None, default_value=self.metadata_filepath)
            log.debug(self.metadata_filepath)
            if os.sep in self.metadata_filepath:
                raise ValueError(f"The provided metadata file {self.metadata_filepath} should be only the name of the metadata file, but it looks like a path.")
            if os.getenv("CONTEXT_MODE") == "TEST":
                self.metadata_filepath = os.path.join(DOCKER_FOLDER_TEST, self.metadata_filepath)
            else:
                self.metadata_filepath = os.path.join(DOCKER_FOLDER_METADATA, self.metadata_filepath)

    def check_parameter(self, key: str, accepted_values: list|None, default_value) -> str | bool | int | None | list:
        try:
            the_parameter = os.getenv(key)
            if the_parameter is None:
                log.error(f"The parameter {key} does not exist as an environment variable. Using default value: {default_value}.")
                return default_value
            elif the_parameter == "":
                log.error(f"The parameter {key} value is empty. Using default value: {default_value}.")
                return default_value
            elif accepted_values is not None:
                if True in accepted_values and False in accepted_values:
                    if the_parameter.lower() == "true":
                        return True
                    elif the_parameter.lower() == "false":
                        return False
                    else:
                        log.error(f"The value '{the_parameter.lower()}' for parameter {key} is not accepted. Using default value: {default_value}.")
                        return default_value
                else:
                    if the_parameter not in accepted_values:
                        log.error(f"The value '{the_parameter.lower()}' for parameter {key} is not accepted. Using default value: {default_value}.")
                        return default_value
                    else:
                        return the_parameter
            else:
                int_parameter = cast_str_to_int(str_value=the_parameter)
                if int_parameter is not None:
                    return int_parameter
                else:
                    # trying to cast as list
                    if the_parameter.startswith("[") and the_parameter.endswith("]"):
                        the_parameter = the_parameter.replace("[", "").replace("]", "")
                        return the_parameter.split(",")
                    else:
                        return the_parameter
        except:
            log.error(f"The parameter {key} does not exist as an environment variable. Using default value: {default_value}.")
            return default_value

    def create_current_working_dir(self):
        # 1. check whether the folder working-dir exists, if not create it
        current_path = os.getcwd()
        working_dir = os.path.join(current_path, WORKING_DIR)
        if not os.path.exists(working_dir):
            log.info(f"Creating the working dir at {working_dir}")
            os.makedirs(working_dir)
        # 2. check whether the db folder exists, if not create it
        working_dir_with_db = os.path.join(working_dir, self.db_name)
        if not os.path.exists(working_dir_with_db):
            log.info(f"Creating a sub-folder for the current database at {working_dir_with_db}")
            os.makedirs(working_dir_with_db)
        # 3. check whether the execution folder exists, if not create it
        execution_folder = os.path.join(working_dir_with_db, self.execution_date.isoformat())
        if not os.path.exists(execution_folder):
            log.info(f"Creating a sub-sub-folder for the current execution at {execution_folder}")
            os.makedirs(execution_folder)
        self.working_dir_current = execution_folder

    def setup_logging_files(self):
        log_file = os.path.join(self.working_dir_current, f"log-{self.execution_date}.log")
        filehandler = logging.FileHandler(log_file, 'a')
        formatter = logging.Formatter('%(asctime)s - %(levelname)s [%(filename)s:%(lineno)d] %(message)s')
        filehandler.setFormatter(formatter)
        setup_logger.log.addHandler(filehandler)  # add the filehandler located in the working dir

    def setup_mapping_to_anonymized_patient_ids(self):
        # pid file should be located at in the DB folder, not the DB+date folder. this allows to:
        # - reuse the same PID file for a database that is consolidated (data is appended) several times
        # - while being sure to reuse the previous PIDs when consolidating the db
        # - we can also easily reset that PID file when db_drop==true (otherwise, we reuse the pid file)
        pid_filename = "anonymized_patient_ids.json"
        if os.getenv("CONTEXT_MODE") == "TEST":
            self.anonymized_patient_ids_filepath = os.path.join(DOCKER_FOLDER_TEST, pid_filename)
        else:
            self.anonymized_patient_ids_filepath = os.path.join(self.working_dir_current, "..", pid_filename)

        log.debug(self.anonymized_patient_ids_filepath)
        if not os.path.exists(self.anonymized_patient_ids_filepath) or os.stat(self.anonymized_patient_ids_filepath).st_size == 0 or self.db_drop is True:
            log.info("write {} in patient ids mapping")
            # the file is empty, we simply add the empty mapping
            # otherwise the file cannot be read as a JSON file
            with open(self.anonymized_patient_ids_filepath, "w") as file:
                file.write("{}")
        else:
            log.info("patient ids mapping already contains data")
            # there are some mappings there, nothing more to do
            pass

    def to_json(self):
        return {"my_exec": "exec"}
        # return dataclasses.asdict(self, dict_factory=factory)

    def __str__(self):
        return json.dumps(self.to_json())



from enums.EnumAsClass import EnumAsClass
from enums.HospitalNames import HospitalNames
from enums.MetadataColumns import MetadataColumns

# cannot use HospitalNames.X in the enum, so I converted it to a dict in
# order to access the vcf columns based on the hospital names
vcf_columns = {
    HospitalNames.RS_IMGGE: MetadataColumns.normalize_name("path_vcf"),
    HospitalNames.ES_HSJD: MetadataColumns.normalize_name("VCF_Path"),
    HospitalNames.IT_BUZZI_UC1: None,
    HospitalNames.IT_BUZZI_UC3: None,
    HospitalNames.ES_TERRASSA: None,
    HospitalNames.DE_UKK: None,
    HospitalNames.ES_LAFE: MetadataColumns.normalize_name("path_vcf"),
    HospitalNames.IL_HMC: None,
    HospitalNames.EXPES_EDA: None,
    HospitalNames.EXPES_COVID: None,
    HospitalNames.EXPES_KIDNEY: None,
    HospitalNames.TEST_H1: None,
    HospitalNames.TEST_H2: None,
    HospitalNames.TEST_H3: None
}
